/* PTS */

package rCr._BUILDER;


import rCr.util.I_Depurador;
import rCr.communicativeAct.C_CommunicativeActHandler;
import rCr.communicativeAct.C_Sensor;
import rCr.communicativeAct.C_Actuator;
import rCr.initiator.I_InitiatorCommunicator;
import rCr.communicator.I_Communicator;
import yab.io.KaConnectOk;
import rCr.util.U_Depurador;
import rCr.util.U_Geral;
import yab.agent.DisasterSpace;
import rCr.agentModel.I_AgentModel;
import rCr.agentWorkingMemory.I_AgentWorkingMemory;
import rCr.agent.skillLevel_0.B_PrimaryCapability;
import rCr.agent.I_Agent;
import rCr.util.U_PreCondicao;
import rCr.util.X_PreCondicao;




/**
 *  @author PTS
 *  @since MAY-2004
 */
abstract public class C_BUILDER_RcR
implements I_BUILDER, I_Depurador
{
   /**
    * @directed
    * @supplierRole initiator 
    */
   private C_CommunicativeActHandler icl_connector = new C_CommunicativeActHandler();
   /**
    * @directed
    * @supplierRole sensor 
    */
   private C_Sensor icl_sensor = new C_Sensor();
   /**
    * @directed
    * @supplierRole actuator
    */
   private C_Actuator icl_actuator = new C_Actuator();

   /**
    * Agent Components
    */
   private I_AgentModel iitf_agentModel = null;
   private I_AgentWorkingMemory iitf_agentWorkingMemory  = null;
   private B_PrimaryCapability icl_primaryCapability = null;
   private I_Agent iitf_agent = null;

   // Agent view of the Disaster Space
   protected DisasterSpace icl_disasterSpace = null;


   /**
    * To be specified by each subclass:
    * - Agent Type
    * - Agent Model
    * - Agent Working Memory
    * - Agent
    **/
   abstract protected int getAgentType();
   abstract protected I_AgentModel createAgentModel();
   abstract protected void setAgentModelCommunicationRestriction( I_AgentModel acl_agentModel );
   abstract protected I_AgentWorkingMemory createAgentWorkingMemory();
   abstract protected I_Agent createAgent();
   abstract protected void postWorkingMemoryInitialize( I_AgentWorkingMemory aitf_agentWorkingMemory );


   public void buildConnector( I_InitiatorCommunicator aitf_initiatorCommunicator )
   throws X_PreCondicao
   {
      U_PreCondicao.assercaoVerdadeira( aitf_initiatorCommunicator, this, "ERROR: aitf_initiatorCommunicator is NULL!" );

      // Obter o objecto comunicador
      //----------------------------
      I_Communicator litf_communicator = ( I_Communicator )aitf_initiatorCommunicator.obterObjectoComunicador();

      // Registar o endereco do processo remoto (mundo no qual o agente existe)
      //-----------------------------------------------------------------------
      litf_communicator.registarEnderecoProcessoRemoto( aitf_initiatorCommunicator.obterEnderecoProcessoRemoto() );

      // Definir o objecto comunicador de cada "CommunicativeActHandler" (Initiator, Sensor, Actuator)
      //----------------------------------------------------------------------------------------------
      icl_connector.setCommunicator( litf_communicator );
      icl_sensor.setCommunicator( litf_communicator );
      icl_actuator.setCommunicator( litf_communicator );
   }


   public void buildAgentToPlatformConnection()
   throws X_PreCondicao
   {
      U_PreCondicao.assercaoVerdadeira( icl_connector, this, "ERROR: icl_connector is NULL!" );

      /**
       * Send connect request & get connect response data
       */
      KaConnectOk lcl_connectionData = icl_connector.akConnect( getAgentType() );
      if( null == lcl_connectionData )
      {
         U_Depurador.escrever( I_Depurador.kint_depurarVerboso,
                               U_Geral.paraNomeClasse( this ),
                                "=== INSUCESSO NA RECEPCAO DE DADOS DA CONEXAO ===" );
         return;
      }
      U_Depurador.escrever( I_Depurador.kint_depurarVerboso,
                            U_Geral.paraNomeClasse( this ),
                             "=== RECEPCAO DOS DADOS DE CONEXAO ===" );

      /**
       * Create "DisasterSpace" with:
       * - agent ID sent by the kernel
       * - geographical data
       */
      icl_disasterSpace = new DisasterSpace( lcl_connectionData.selfId,
                                             lcl_connectionData.selfAndMap);

      // Send a connect acknowledge ("KaConnectOK") to the kernel
      // --------------------------------------------------------
      icl_connector.akAcknowledge( icl_disasterSpace.self.id );

      U_Depurador.escrever( I_Depurador.kint_depurarInfo,
                            U_Geral.paraNomeClasse( this ),
                            icl_disasterSpace.self.toString() );
   }


   public void buildAgentModel()
   throws X_PreCondicao
   {
      U_PreCondicao.assercaoVerdadeira( icl_sensor, this, "ERROR: icl_sensor is NULL!" );

      // Create the Agent Model (Architecture Implementation)
      // (RcR default is: rCr.agentModel.C_RcR)
      // ----------------------------------------------------
      iitf_agentModel = createAgentModel();
      iitf_agentModel.setSensor( icl_sensor );
      setAgentModelCommunicationRestriction( iitf_agentModel );
   }


   public void buildAgentWorkingMemory()
   throws X_PreCondicao
   {
      U_PreCondicao.assercaoVerdadeira( icl_disasterSpace, this, "ERROR: icl_disasterSpace is NULL!" );

      // Create the Agent's Working Memory
      // (RcR default is: rCr.agentWorkingMemory.E_WorkingMemory_RcR)
      // ------------------------------------------------------------
      iitf_agentWorkingMemory = createAgentWorkingMemory();
      iitf_agentWorkingMemory.initialize( icl_disasterSpace );
      postWorkingMemoryInitialize( iitf_agentWorkingMemory );
   }


   public void buildAgentPrimaryCapability()
   throws X_PreCondicao
   {
      U_PreCondicao.assercaoVerdadeira( icl_actuator, this, "ERROR: icl_actuator is NULL!" );

      // Set the Agent Primary Capabilities (action and interaction)
      // Register them in the "Agent" and in the "Agent Model"
      // -----------------------------------------------------------
      icl_primaryCapability = new B_PrimaryCapability( icl_actuator );
   }


   public void buildAgent()
   {
      // Create the Agent
      // ----------------
      iitf_agent = createAgent();
   }


   public void buildAgentToComponentLinkage()
   throws X_PreCondicao
   {
      U_PreCondicao.assercaoVerdadeira( iitf_agentModel, this, "ERROR: iitf_agentModel is NULL!" );
      U_PreCondicao.assercaoVerdadeira( iitf_agent, this, "ERROR: iitf_agent is NULL!" );
      U_PreCondicao.assercaoVerdadeira( iitf_agentWorkingMemory, this, "ERROR: iitf_agentWorkingMemory is NULL!" );
      U_PreCondicao.assercaoVerdadeira( icl_primaryCapability, this, "ERROR: icl_primaryCapability is NULL!" );

      // Set the "Agent" that the "Agent Model" controls
      // -----------------------------------------------
      iitf_agentModel.setAgent( iitf_agent );

      // Set the "Working Memory" of the "Agent"
      // ---------------------------------------
      iitf_agent.setAgentWorkingMemory( iitf_agentWorkingMemory );

      // Set the "Primary Capability" of the "Agent" (and "Agent Model")
      // ---------------------------------------------------------------
      iitf_agent.setPrimaryCapability( icl_primaryCapability );
      iitf_agentModel.setPrimaryCapability( icl_primaryCapability );

   }


   public Runnable getResult()
   {
      if( null == iitf_agentModel )
      {
         throw new Error( "ERROR: iitf_agentModel is NULL!" );
      }

      return ( Runnable )iitf_agentModel;
   }
}
