/* PTS */


package rCr.agent.rCrAgentRRR;


import rCr.agentWorkingMemory.E_WorkingMemory_RRR;
import yab.agent.DisasterSpace;
import rCr.agentWorkingMemory.I_AgentWorkingMemory;
import yab.agent.object.RealObject;
import yab.io.KaSense;
import yab.agent.object.Humanoid;
import rCr.agent.skillLevel_0.B_PrimaryCapability;
import rCr.agent.skillLevel_1.I_Planner;
import rCr.agent.skillLevel_1.I_PlannerBaseline;
import rCr.agent.skillLevel_1.B_GIS_Router_Yab;
import rCr.agent.I_Agent;
import java.util.Iterator;
import yab.agent.object.FireBrigade;
import yab.agent.object.AmbulanceTeam;
import yab.agent.object.PoliceForce;
import yab.agent.object.Civilian;
import rCr.agentWorkingMemory.E_SpaceGrid;
import rCr.agent.skillLevel_0.I_PrimaryAction;
import yab.agent.object.PoliceOffice;
import yab.agent.object.AmbulanceCenter;
import yab.agent.object.FireStation;
import rCr.agent.skillLevel_0.I_PrimaryInteraction_Constants;


/**
 *  @author PTS
 *  @since JUN.2004
 */
abstract public class C_Agent
implements I_Agent, I_PrimaryInteraction_Constants
{
   // --------------------------------------
   // The Agent Working Memory RRR includes:
   // - current disaster space
   // - space grid partition
   // --------------------------------------
   public E_WorkingMemory_RRR workingMemory = null;
   // -----------------------
   // "To Simplify" variables
   // -----------------------
   // To simplify: "my" Institution MapGrid Coordination Perspective (also within WorkingMemory)
   protected E_SpaceGrid mySpaceGrid = null;
   // To simplify: current simulation time (also within currentDisasterSpace)
   protected int currentTime = 0;
   // To simplify "disasterSpace" access (also within WorkingMemory)
   protected DisasterSpace currentDisasterSpace = null;
   // -----------------------

   // The planning capability (with several routing strategies)
   protected I_Planner plannerCapability = null;
   // Agent Age (ID order at its Collection)
   private Integer iInt_Age = null;
   // Vision and Voice Hear Radius (10 and 30meters) [see RcR2004 Rules]
   protected int pint_visionRadius = 10000;
   protected int pint_hearRadius = 30000;


   // Invoked by the utility method:
   // - handleCommunicativeAct( int ID, ...)
   // (SubClasses deal with RealObject instead of ID)
   //abstract protected void handleCommunicativeAct( RealObject aobj_sender, String astr_content );
   abstract protected void handleHearSay( RealObject acl_sender, String astr_content );
   abstract protected void handleHearTell( RealObject acl_sender, String astr_content );


   // ============================================
   // (BEGIN) type cast abstract redefined methods
   // ============================================
   // Agent Primary Capability
   abstract public void setAgentPrimaryCapability( B_PrimaryCapability acl_primaryCapability );
   // Each "self" RealObject SubClass
   private RealObject self()
   { return ( RealObject )currentDisasterSpace.self; }
   // ==========================================
   // (END) type cast abstract redefined methods
   // ==========================================


   /**
    * Age:
    * ----
    * The Agent ID order at its DisasterSpace Collection
    * Younger = 1
    * Older = Collection.size()
    * Age = {1, 2, ..., Collection.size()}
    */
   protected int getAge()
   {
      // If Age already computed
      // then return that value
      if( iInt_Age != null )
      {
         return iInt_Age.intValue();
      }

      // Get the Collection to iterate
      Iterator litf_iterator = null;
      if( currentDisasterSpace.self instanceof FireBrigade )
      {
         litf_iterator = currentDisasterSpace.fireBrigades.iterator();
      }
      if( currentDisasterSpace.self instanceof AmbulanceTeam )
      {
         litf_iterator = currentDisasterSpace.ambulanceTeams.iterator();
      }
      if( currentDisasterSpace.self instanceof PoliceForce )
      {
         litf_iterator = currentDisasterSpace.policeForces.iterator();
      }
      if( currentDisasterSpace.self instanceof Civilian )
      {
         int lint_civilianAge = 1;
         println( "====== Civilian Detected: for testing purpose only ====== ( age set to " + lint_civilianAge + " )" );
         return lint_civilianAge;
      }
      if( litf_iterator == null )
      {
         throw new Error( "PTS: getAge:: valid Collection expected" );
      }
      int lint_selfId = self().id;
      int lint_age = 0;
      for( ; litf_iterator.hasNext(); )
      {
         Humanoid lcl_humanoid = ( Humanoid )litf_iterator.next();
         if( ( lcl_humanoid != null ) &&
             ( lcl_humanoid.id <= lint_selfId  ) )
         {
            ++lint_age;
         }
      }
      if( lint_age == 0 )
      {
         throw new Error( "PTS: getAge:: age > 0 expected" );
      }
      // Register and return the computed Age
      iInt_Age = new Integer( lint_age );
      return lint_age;
   }


   public void setAgentWorkingMemory( I_AgentWorkingMemory agentWorkingMemory )
   {
      // Set WorkingMemory
      this.workingMemory = ( E_WorkingMemory_RRR )agentWorkingMemory;
      // Set DisasterSpace
      this.currentDisasterSpace = this.workingMemory.getCurrentDisasterSpace();
      // If there is any empty SpaceGrid
      // then build a "default Space Grid"
      // (default values defined at I_AgentWorkingMemory: 1x1 Grid)
      buildAnyEmptySpaceGrid();
      // Set My (self agent) SpaceGrid
      setMySpaceGrid();
   }


   public void setPrimaryCapability( B_PrimaryCapability acl_primaryCapability )
   {
      acl_primaryCapability.setAgentID( self().id );
      setAgentPrimaryCapability( acl_primaryCapability );
   }


   public void setHigherCapability( I_Planner acl_plannerCapability )
   {
      this.plannerCapability = acl_plannerCapability;
      this.plannerCapability.setAgentSelf( self() );
   }


   public void setHigherCapability( I_PlannerBaseline acl_plannerCapability )
   {
      throw new UnsupportedOperationException( "PTS: RRR AGENT ONLY SUPORTS I_Planner CAPABILITY" );
   }


   public boolean isSelf ( int aint_agentID )
   {
      return self() == currentDisasterSpace.get( aint_agentID );
   }


   public void preSimulation()
   {
   }


   public void updateWorld( Object aobj_percepcao )
   {
      KaSense lobj_kaSense = ( KaSense )aobj_percepcao;
      // Update Simulation Time (the "to simplify" variable)
      currentTime = lobj_kaSense.time;
      // Update currentDisasterSpace
      currentDisasterSpace.update( lobj_kaSense.selfAndMap, currentTime );
   }


   /**
    * Handle a communicative act sent by "aint_senderID" with content "astr_content"
    */
   public void handleHearSay( int aint_senderID, String astr_content )
   {
      RealObject lobj_sender = currentDisasterSpace.get( aint_senderID );
      handleHearSay( lobj_sender, astr_content );
   }


   /**
    * Handle a communicative act sent by "aint_senderID" with content "astr_content"
    */
   public void handleHearTell( int aint_senderID, String astr_content )
   {
      RealObject lobj_sender = currentDisasterSpace.get( aint_senderID );
      handleHearTell( lobj_sender, astr_content );
   }


   public boolean acceptHearSay( int aint_senderID, int aint_remainingHearingUtterances )
   {
      return true;
   }


   public boolean acceptHearTell( int aint_senderID, int aint_remainingHearingUtterances )
   {
      return true;
   }


   private void buildAnyEmptySpaceGrid()
   {
      workingMemory.buildAnyEmptySpaceGrid( currentDisasterSpace );
      if( self() instanceof Civilian )
      {
         workingMemory.buildEmptySpaceGrid_CIVILIAN( currentDisasterSpace );
      }
   }


   private void setMySpaceGrid()
   {
      if( ( self() instanceof PoliceForce ) ||
          ( self() instanceof PoliceOffice ) )
      {
         this.mySpaceGrid = workingMemory.getSpaceGrid_POLICE();
         return;
      }

      if( ( self() instanceof AmbulanceTeam ) ||
          ( self() instanceof AmbulanceCenter ) )
      {
         this.mySpaceGrid = workingMemory.getSpaceGrid_AMBULANCE();
         return;
      }

      if( ( self() instanceof FireBrigade ) ||
          ( self() instanceof FireStation ) )
      {
         this.mySpaceGrid = workingMemory.getSpaceGrid_FIRE();
         return;
      }

      if( self() instanceof Civilian )
      {
         this.mySpaceGrid = workingMemory.getSpaceGrid_CIVILIAN();
         return;
      }
   }


   private void println( String astr_message )
   {
      System.out.println( astr_message );
   }
}
