/* PTS */


package rCr.agent.rCrAgentRRR.r0;


import rCr.agentWorkingMemory.E_WorkingMemory;
import java.beans.Statement;
import yab.agent.DisasterSpace;
import yab.agent.object.RealObject;
import rCr.util.X_RcR_Accao;
import rCr.util.U_LeitorEscritorLinha;
import rCr.agent.skillLevel_0.I_PrimaryInteraction;
import rCr.agent.skillLevel_0.I_PrimaryAction;
import rCr.agent.skillLevel_0.I_PrimaryCapability_PoliceForce;
import rCr.agent.skillLevel_0.I_PrimaryCapability_AmbulanceTeam;
import rCr.agent.skillLevel_0.I_PrimaryCapability_FireBrigade;
import rCr.agentWorkingMemory.E_FactMemory;
import rCr.agentWorkingMemory.E_FactMemory.E_Fact;
import rCr.agentWorkingMemory.I_FactMemoryConstants;
import rCr.util.U_Geral;


/**
 *  @author PTS
 *  @since JUN.2004
 */
public class E_Rule
{
   /**
    * Working Memory of this Rule
    * (with access to the Agent WorkingMemory)
    * Contains:s
    * - the set of "candidateStatementRuleArguments"
    * - the set of "executedStatementRule"
    */
   private E_FactMemory icl_factMemory = null;
   /**
    * OPTIONAL: PreCondition to be evaluated before "icl_situationToFireRule"
    * It is defined with "setPreCondition" method
    * It is not defined with class the constructor
    * (enables to define restrictions that are independent of a particular "situation")
    * (e.g. of a situation is the HEALTH_SITUATION [see .xls file])
    * @supplierRole preCondition 
    */
   private E_ConditionRule icl_preCondition = null;
   /**
    * Situation to be satisfied in order to fire the Rule
    * @supplierRole situationToFireRule 
    */
   private E_ConditionRule icl_situationToFireRule = null;
   /**
    * Statement to be issued when the Rule fires
    */
   private E_StatementRule icl_statementRule = null;
   /**
    * Salience
    * higher value ==>> lower priority
    * (to read like: 1st 2nd 4rd 4th 5th ...)
    * default value == the last rule to get evaluated
    */
   private int iint_salience = Integer.MAX_VALUE;


   public E_Rule( E_WorkingMemory acl_workingMemory, E_ConditionRule acl_condition, E_StatementRule acl_statementRule )
   {
      this.icl_factMemory = new E_FactMemory( acl_workingMemory );
      this.icl_situationToFireRule = acl_condition;
      this.icl_statementRule = acl_statementRule;
   }


   public boolean isReadyToFire()
   {
      boolean lbln_isReadyToFire = false;
      boolean lbln_mustEvalPreCondition = ( icl_preCondition != null );
      boolean lbln_isPossibleToEvalSituationToFireRule = true;
      // If there is a PreCondition, evaluate it
      if( lbln_mustEvalPreCondition )
      {
         lbln_isPossibleToEvalSituationToFireRule = icl_preCondition.eval( icl_factMemory );
      }

      // If it is possible, evaluate the condition to fire rule
      if( lbln_isPossibleToEvalSituationToFireRule )
      {
         lbln_isReadyToFire = icl_situationToFireRule.eval( icl_factMemory);
      }
      return lbln_isReadyToFire;
   }


   public void fire()
   throws X_RcR_Accao
   {
      // --------------------------------
      // Get the Statement to be executed
      // --------------------------------
      Statement lstt_statement = icl_statementRule.getStatement( icl_factMemory );
      // ------------------------------------------------------
      // Assert Statement to be executed both in:
      // - this FactMemory
      // - the AgentFactMemory
      // (all statements are registered at the AgentFactMemory)
      // ------------------------------------------------------
      E_Fact lcl_fact = new E_Fact( I_FactMemoryConstants.kint_executedStatementRule,
                                    lstt_statement );
      icl_factMemory.assertFact( lcl_fact );
      icl_factMemory.getAgentWorkingMemory().getFactMemory().assertFact( lcl_fact );
      try
      {
         // -----------------
         // Execute Statement
         // -----------------
         // println( lstt_statement );
         lstt_statement.execute();
      }
      catch( X_RcR_Accao lcl_exception )
      {
         throw lcl_exception;
      }
      catch( Exception lcl_exception )
      {
         U_LeitorEscritorLinha.escrever( "PTS: Problem with primaryCapability execution" );
         U_LeitorEscritorLinha.escrever( lcl_exception );
      }
   }


   public void setPreCondition( E_ConditionRule acl_preCondition )
   {
      this.icl_preCondition = acl_preCondition;
   }


   public boolean isCommunicationStatement()
   {
      String lstr_methodName = icl_statementRule.getStatement( icl_factMemory ).getMethodName();
      return ( lstr_methodName.equals( I_PrimaryInteraction.kstr_say ) ||
               lstr_methodName.equals( I_PrimaryInteraction.kstr_tell ) );
   }


   public boolean isActionStatement()
   {
      String lstr_methodName = icl_statementRule.getStatement( icl_factMemory ).getMethodName();
      return ( lstr_methodName.equals( I_PrimaryAction.kstr_rest ) ||
               lstr_methodName.equals( I_PrimaryAction.kstr_move ) ||
               lstr_methodName.equals( I_PrimaryCapability_PoliceForce.kstr_clear ) ||
               lstr_methodName.equals( I_PrimaryCapability_AmbulanceTeam.kstr_load ) ||
               lstr_methodName.equals( I_PrimaryCapability_AmbulanceTeam.kstr_unload ) ||
               lstr_methodName.equals( I_PrimaryCapability_AmbulanceTeam.kstr_rescue ) ||
               lstr_methodName.equals( I_PrimaryCapability_FireBrigade.kstr_extinguish ) );
   }


   public void setSalience( int aint_salience )
   {
      this.iint_salience = aint_salience;
   }


   public int getSalience()
   {
      return iint_salience;
   }


   public String getStatementName()
   {
      return icl_statementRule.getStatement( icl_factMemory ).getMethodName();
   }


   public Object[] getStatementArguments()
   {
      return icl_statementRule.getStatement( icl_factMemory ).getArguments();
   }


   private void println( String astr_message )
   {
	   System.out.println( astr_message );
	}


   private void println( Statement astt_statement )
   {
      println( ">> PTS >> " + U_Geral.toString( astt_statement ) );
   }
}