/* PTS */


package rCr.agent.rCrAgentRRR.r1;


import java.util.ArrayList;
import yab.agent.object.MotionlessObject;
import yab.agent.object.MovingObject;
import java.util.Iterator;
import yab.agent.object.Road;
import rCr.agent.skillLevel_1.I_Planner;
import rCr.agent.skillLevel_1.B_GIS_Router_FromTo;
import java.util.Collections;
import rCr.util.U_LeitorEscritorLinha;
import rCr.util.U_Geral;
import java.util.Collection;


/**
 *  @author PTS
 *  @since JUN.2004
 */
public class C_PlanMonitor_Route
{
   /**
    * A configurable constant that indicates
    * the expected progress at each time slot (simulaiton cycle)
    * @see RcR manual p.11 (Maximum speed)
    * The maximum speed of a platoon is 20 [km/h],
    * so a platoon can move 333m in one cycle at most.
    * Actually, a humanoid seldom moves at the maximum speed through one cycle,
    * because a humanoid has to pass cross-points, change lanes, avoid blockades, etc.
    * (333 meters == 333000 mm)
    * PTS: this value is set a litle lower:
    * (to 300000 == 300 meters per cycle)
    */
   static private final int pint_maximumExpectedProgressAtEachTime = 300000;
   /**
    * A configurable constant that indicates the admissible delay
    *
    * at each time cycle
    * PTS: this value default is set to 1/3 of kint_maximumExpectedProgressAtEachTime
    */
   private int pint_admissibleDelay = pint_maximumExpectedProgressAtEachTime / 3;
   /**
    * Agent Object to Monitor
    * - a MovingObject (from DisasterSpace)
    * (defined once at Constructor)
    */
   private final MovingObject icl_agentObjectToMonitor;
   /**
    * Monitor Time information
    * - ON time instant
    * - current time instant
    */
   private int iint_timeON = -1;
   private int iint_timeCurrent = -1;
   /**
    * Plan (a Route) to Monitor during its execution
    * - each element at ArrayList is a MotionlessObject (from DisasterSpace)
    */
   private ArrayList iarr_planToMonitor = null;
   /**
    * The list of taboo locations detected during "repairPlan" execution
    */
   private ArrayList iarr_tabooLocationSetHistory = new ArrayList();
   /**
    * Counts the number of plan repairs
    */
   private int iint_totalNumberOfPlanRepair = 0;
   /**
    * Location (MotionlessObject) information
    * - location at current time
    * - location expected at next time
    * Note: -1 indicates a "location out of plan"
    */
   private int iint_indexOfCurrentLocation = -1;
   private int iint_indexOfExpectedLocationAtNextTime = -1;
   /**
    * The accumulated deviation from "timeON"
    * (updated at each "setTime" method invocation)
    */
   private int iint_accumulatedDeviation = 0;
   /**
    * Unexpected cost detected
    * Each cost unit is a simulation cycle
    */
   private int iint_unexpectedCost = 0;


   public C_PlanMonitor_Route( MovingObject acl_agentObjectToMonitor )
   {
      this.icl_agentObjectToMonitor = acl_agentObjectToMonitor;
   }


   public void setON( int aint_timeCurrent, ArrayList aarr_planToMonitor )
   {
      // Monitor Reset
      setOFF();
      if( aarr_planToMonitor.isEmpty() )
      {
         return;
      }
      // Monitor Initialization
      this.iint_timeON = aint_timeCurrent;
      this.iint_timeCurrent = iint_timeON;
      this.iarr_planToMonitor = aarr_planToMonitor;
      this.iarr_tabooLocationSetHistory = new ArrayList();
      this.iint_totalNumberOfPlanRepair = 0;
      this.iint_accumulatedDeviation = 0;
      this.iint_unexpectedCost = 0;
      updateIndexOfCurrentLocation();
      updateIndexOfExpectedLocationAtNextTime ();
   }


   /**
    * Return true if:
    * - monitor is ON and
    * - the plan is not finished
    */
   public boolean isExecuting()
   {
      return ( existPlanToMonitor() && ( ! isPlanConcluded () ) );
   }


   public void updateTime( int aint_time )
   {
      if( ! existPlanToMonitor() )
      {
         throw new Error( "PTS: updateTime - plan ON expected" );
      }
      int lint_elapsedTime = aint_time - iint_timeCurrent;
      if( lint_elapsedTime < 0 )
      {
         throw new Error( "PTS: updateTime - time progress expected" );
      }
      // Note: ( elapsedTime == 0 )
      // do not return immediately
      // it is importanto to allow execution of this method
      // (to dela with the out of plan situation)

      // If plan already executed
      // then do nothing
      if( isPlanConcluded() )
      {
         return;
      }
      // Update Time
      // -----------
      iint_timeCurrent = aint_time;

      // Now current time become previous time (time advences!)
      // keep previous information:
      // - index of location
      // - execution out of plan
      int lint_indexOfPreviousLocation = iint_indexOfCurrentLocation;
      boolean lbln_previousExecutionOutOfPlan = isExecutionOutOfPlan();
      // Update Index of Current Location
      updateIndexOfCurrentLocation();
      // -----
      // Note:
      // -----
      // If execution gets (or stays) out of the plan
      // then just return
      // (maybe next time (later) agent will retake this plan)
      // Note: "getRemaining" Plan repairs the plan if location is out of plan
      if( isExecutionOutOfPlan() )
      {
         return;
      }
      // If previous execution was out of the plan (lbln_previousExecutionOutOfPlan)
      // (has just returned back to the plan)
      // then just accumlate the respective deviation (set effectiveProgress = 0)
      int lint_effectiveProgress = 0;
      if( ! lbln_previousExecutionOutOfPlan )
      {
         lint_effectiveProgress = distanceFromTo(
            lint_indexOfPreviousLocation, iint_indexOfCurrentLocation, iarr_planToMonitor  );
      }
      // -------------------------------
      // Update the acumulated deviation
      // -------------------------------
      int lint_elapsedTimeToConsider = Math.max( lint_elapsedTime - iint_unexpectedCost, 0 );
      iint_unexpectedCost = Math.max( iint_unexpectedCost - lint_elapsedTime, 0 );
      int lint_maximumExpectedProgress = lint_elapsedTimeToConsider * pint_maximumExpectedProgressAtEachTime;
      int lint_deviation = lint_maximumExpectedProgress - lint_effectiveProgress;
      iint_accumulatedDeviation = iint_accumulatedDeviation + lint_deviation;

      // If plan execution is not finished
      // then update index of "expected location at next time"
      if( ! isPlanConcluded () )
      {
         updateIndexOfExpectedLocationAtNextTime();
      }
      else
      {
         iint_indexOfExpectedLocationAtNextTime = iint_indexOfCurrentLocation;
      }
   }


   /**
    * Update the unexpected cost
    */
   public void acceptUnexpectedCost( int aint_unexpectedCost )
   {
      this.iint_unexpectedCost = aint_unexpectedCost;
   }


   public void setOFF()
   {
      this.iarr_planToMonitor = null;
      this.iarr_tabooLocationSetHistory = new ArrayList();
      this.iint_totalNumberOfPlanRepair = 0;
      this.iint_timeON = -1;
      this.iint_timeCurrent = -1;
      this.iint_indexOfCurrentLocation = -1;
      this.iint_indexOfExpectedLocationAtNextTime= -1;
      this.iint_accumulatedDeviation = 0;
      this.iint_unexpectedCost = 0;
   }


   public MotionlessObject getPlanDestination()
   {
      if( ! existPlanToMonitor () )
      {
         return null;
      }
      int lint_lastIndex = iarr_planToMonitor.size() - 1;
      MotionlessObject lcl_planDestination = ( MotionlessObject )iarr_planToMonitor.get( lint_lastIndex );
      return lcl_planDestination;
   }


   public boolean isEqualDestination( MotionlessObject acl_destination )
   {
      MotionlessObject lcl_planDestination = getPlanDestination();
      if( lcl_planDestination == null )
      {
         return false;
      }
      return ( acl_destination == lcl_planDestination );
   }


   public boolean isExecutionOutOfPlan()
   {
      if( isExecuting() )
      {
         return ( iint_indexOfCurrentLocation < 0 );
      }
      return false;
   }


   /**
    * Return 0 (zero) if the "planIsConcluded" (nothing to repair)
    * else return the evaluated need "K" to repair this plan
    * K = accumulatedDeviation / (maximumExpectedProgressAtEachTim + admissibleDelay)
    * ----
    * K = 0 ==>> no need (progress according to plan)
    * K < 0 ==>> K admissibleDelay violations
    * K > 0 ==>> K admissibleDelay extra performance
    * ----
    * Note: this velue refers the current plan,
    *       even if "isPlanExecuted == true"
    */
   public int getNeedToRepairPlan()
   {
      if( isPlanConcluded() )
      {
         return 0;
      }
      return iint_accumulatedDeviation / ( pint_maximumExpectedProgressAtEachTime + pint_admissibleDelay);
   }


   public int getTotalNumberOfPlanRepair()
   {
      return iint_totalNumberOfPlanRepair;
   }


   /**
    * Utility method
    * Repair the current plan
    * Using the default RepairStrategy
    * Initialize the Monitor with the new "repair plan"
    * @see "repairPlan" with arity 2
    * @see reagarding RepairStrategy: "buildRepairPlan" with arity 2
    */
   public boolean repairPlan( int aint_time )
   {
      return repairPlan( aint_time, null );
   }


   /**
    * Repair the current plan
    * Use the intended RepairStrategy
    * @return true if a new plan was built (and monitor is now acting on it)
    * @return false is an empty plan was built (and monitor is setOFF)
    * @see reagarding RepairStrategy: "buildRepairPlan" with arity 2
    * Initialize the Monitor with the new "repair plan"
    * Keeps the old taboo location set
    */
   public boolean repairPlan( int aint_time, E_RepairStrategy_FromTo aC_intendedRepairStrategy )
   {
      final boolean lbln_newPlanWasBuilt = true;
      ArrayList larr_repairPlan = buildRepairPlan( aint_time, aC_intendedRepairStrategy);
      if( larr_repairPlan.isEmpty() )
      {
         setOFF();
         return ( ! lbln_newPlanWasBuilt );
      }

      ArrayList larr_oldTabooLocaionListSetHistory = this.iarr_tabooLocationSetHistory;
      int lint_oldTotalPlanRepairExecution = this.iint_totalNumberOfPlanRepair;
      setON( aint_time, larr_repairPlan );
      this.iarr_tabooLocationSetHistory = larr_oldTabooLocaionListSetHistory;
      this.iint_totalNumberOfPlanRepair = lint_oldTotalPlanRepairExecution + 1;
      return lbln_newPlanWasBuilt;
   }


   /**
    * Utility method
    * When executionOutOfPlan, Repair the current plan
    * using the default RepairStrategy
    * @see "getRemaningPlan" with arity 2
    * @see reagarding RepairStrategy: "buildRepairPlan" with arity 2
    */
   public ArrayList getRemaningPlan( int aint_time )
   {
      return getRemaningPlan( aint_time, null );
   }


   /**
    * Repair the current plan
    * Use the intended RepairStrategy
    * @see reagarding RepairStrategy: "buildRepairPlan" with arity 2
    */
   public ArrayList getRemaningPlan( int aint_time, E_RepairStrategy_FromTo aC_intendedRepairStrategy )
   {
      // 1st update time
      // ---------------
      updateTime( aint_time );

      // An initial empty plan
      ArrayList larr_remaningPlan = new ArrayList();
      if(  isPlanConcluded () )
      {
         return larr_remaningPlan;
      }
      // Automatic repair of an out of plan situaion
      // --------------------------------------------
      if( isExecutionOutOfPlan() )
      {
         larr_remaningPlan = buildRepairPlanWhenExecutionOutOfPlan( aC_intendedRepairStrategy );
         iarr_planToMonitor = larr_remaningPlan;
         updateIndexOfCurrentLocation();
         updateIndexOfExpectedLocationAtNextTime ();
         return larr_remaningPlan;
      }

      // Build a plan that begins at "indexOfCurrentLocation"
      // ----------------------------------------------------
      for( int lint_iLoop = iint_indexOfCurrentLocation;
           lint_iLoop < iarr_planToMonitor.size();
           lint_iLoop++ )
      {
         larr_remaningPlan.add( iarr_planToMonitor.get( lint_iLoop ) );
      }
      return larr_remaningPlan;
   }


   public MotionlessObject getExpectedLocationAtNextTime()
   {
      if( iint_indexOfExpectedLocationAtNextTime < 0 )
      {
         return null;
      }
      return ( MotionlessObject )iarr_planToMonitor.get( iint_indexOfExpectedLocationAtNextTime );
   }


   public int getEpasedTime()
   {
      if( ! existPlanToMonitor () )
      {
         throw new Error( "PTS: getEpasedTime - plan ON expected" );
      }
      return iint_timeCurrent - iint_timeON;
   }

   /**
    * Return the expected time (number of cycles)
    * needed to execute the "planToEval"
    */
   public static int getExpectedPlanExecutionTime( ArrayList aarr_planToEval )
   {
      int lint_totalDistance = distanceFromTo( 0, aarr_planToEval.size() - 1, aarr_planToEval );
      return ( int )Math.ceil( lint_totalDistance / pint_maximumExpectedProgressAtEachTime);
   }


   /**
    * Return the distance [mm] between "from" and "to" indexes
    * @param aarr_route ArrayList of MotionlessObject
    */
   static public int distanceFromTo( int aint_indexFrom, int aint_indexTo, ArrayList aarr_route )
   {
      int lint_from = ( aint_indexFrom < aint_indexTo ) ? aint_indexFrom : aint_indexTo;
      int lint_to = ( aint_indexFrom < aint_indexTo ) ? aint_indexTo : aint_indexFrom;
      int lint_index = lint_from;
      int lint_distanceSum = 0;
      // If both equal then distance = 0
      if( lint_from == lint_to )
      {
         return lint_distanceSum;
      }
      // Note: sum the length of the last ("to") index (if Road)
      while( ( lint_index <= lint_to ) &&
             ( lint_index < aarr_route.size() ) &&
             ( lint_index >= 0 ) )
      {
         MotionlessObject lcl_location = ( MotionlessObject )aarr_route.get( lint_index );
         if( lcl_location instanceof Road )
         {
            lint_distanceSum = lint_distanceSum + ( ( Road )lcl_location ).length();
         }
         ++lint_index;
      }
      return lint_distanceSum;
   }


   /**
    * Return true if:
    * - the plan is finished
    * @throws Error exception if monitor is OFF
    */
   private boolean isPlanConcluded()
   {
      if( ! existPlanToMonitor() )
      {
         throw new Error( "PTS: isPlanConcluded - plan ON expected" );
      }
      int lint_lastIndex = iarr_planToMonitor.size() - 1;
      return ( iint_indexOfCurrentLocation == lint_lastIndex );
   }


   /**
    * Return true if:
    * - exist a plan to monitor
    */
   private boolean existPlanToMonitor()
   {
      return ( iarr_planToMonitor != null );
   }


   /**
    * ----------------------------------------------
    * Utility method (uses a default RepairStrategy)
    * To use other RepairStrategy
    * implement E_RepairStrategy_FromTo
    * (and used buildRepairPlan with arity 2)
    * ----------------------------------------------
    * @see buildRepairPan / arity 2
    * @see buildDefaultRepairStrategy_FromTo
    */
   private ArrayList buildRepairPlan( int aint_time )
   {
      return buildRepairPlan( aint_time, null );
   }

   /**
    * Repair Plan (with intended RepairStrategy
    * (other must be implmented at subclasses)
    * This Repair Plan is built as:
    * - a route using the intended RepairStrategy
    *   default is given by "buildDefaultRepairStrategy_FromTo")
    *   - to the same destination avoiding the next location in the plan
    * - an empty route if:
    *   - "isExecutionOutOfPlan" or
    *   - "isPlanExecutedAndMonitorON" or
    *   - "indexOfCurrentLocation == 0" and it has no appliable neighbours
    */
   private ArrayList buildRepairPlan( int aint_time, E_RepairStrategy_FromTo aC_intendedRepairStrategy )
   {
      // 1st update time
      // ---------------
      updateTime( aint_time );

      // An initial empty plan
      ArrayList larr_repairPlan = new ArrayList();
      if(  isPlanConcluded () )
      {
         return larr_repairPlan;
      }
      if( isExecutionOutOfPlan() )
      {
         return buildRepairPlanWhenExecutionOutOfPlan( aC_intendedRepairStrategy );
      }
      // Build the Repair Plan
      // - destination (the plan destination)
      // - origin (the previous location)
      // - taboo (the current location)
      // ------------------------------------
      // Destination:
      // - the same as the last plan index
      int lint_lastIndex = iarr_planToMonitor.size() - 1;
      MotionlessObject lcl_destination = ( MotionlessObject )iarr_planToMonitor.get( lint_lastIndex );
      MotionlessObject lcl_newTabooLocation = ( MotionlessObject )iarr_planToMonitor.get( iint_indexOfCurrentLocation );

      // Origin:
      // If there is not a previous location in the plan
      // then get a neighbour of "indexOfCurrentLocation+1" location other than:
      // - the newTabooLocation and any other previously detected tabooLocations
      int lint_firstIndex = iint_indexOfCurrentLocation - 1;
      MotionlessObject lcl_origin = null;
      if( lint_firstIndex < 0 )
      {
         MotionlessObject lcl_notReachableNeighbour = ( MotionlessObject )iarr_planToMonitor.get( iint_indexOfCurrentLocation + 1 );
         boolean lbln_found = false;
         for( Iterator litf_iterator = lcl_newTabooLocation.neighborhood().iterator();
              litf_iterator.hasNext() && ( ! lbln_found ) ; )
         {
            // Origin: neighbour other than:
            // - the new taboo location (location at indexOfCurrentLocation)
            // - any previously detected tabooLocation
            lcl_origin = ( MotionlessObject )litf_iterator.next();
            lbln_found = ( lcl_origin != lcl_newTabooLocation ) &&
                         ( lcl_origin != lcl_notReachableNeighbour ) &&
                         ( ! iarr_tabooLocationSetHistory.contains( lcl_origin ) );
         }
         if( ! lbln_found )
         {
            return larr_repairPlan;
         }
      }
      else
      {
         // Origin: the previous location in the plan
         lcl_origin = ( MotionlessObject )iarr_planToMonitor.get( lint_firstIndex );
      }
      // New Taboo Location:
      // - add to taboo location set
      iarr_tabooLocationSetHistory.add( lcl_newTabooLocation );

      // Build the Repair Plan
      // ---------------------
      // Use:
      // - the intended RepairStrategy (if != null)
      // - the "default" RepairStrategy (implemented at "buildDefaultRepairStrategy_FromTo")
      E_RepairStrategy_FromTo lC_repairStrategy = aC_intendedRepairStrategy;
      if( lC_repairStrategy == null )
      {
         lC_repairStrategy = buildDefaultRepairStrategy_FromTo();
      }
      larr_repairPlan = lC_repairStrategy.eval( lcl_origin, lcl_destination, iarr_tabooLocationSetHistory );

      // Add the location at iint_indexOfCurrentLocation to the top of the repairPlan
      larr_repairPlan.add( 0, iarr_planToMonitor.get( iint_indexOfCurrentLocation ) );
      return larr_repairPlan;
   }


   private ArrayList buildRepairPlanWhenExecutionOutOfPlan( E_RepairStrategy_FromTo aC_intendedRepairStrategy )
   {
      MotionlessObject lcl_origin = icl_agentObjectToMonitor.motionlessPosition();
      int lint_lastIndex = iarr_planToMonitor.size() - 1;
      MotionlessObject lcl_destination = ( MotionlessObject )iarr_planToMonitor.get( lint_lastIndex );
      // Build the Repair Plan
      // ---------------------
      // Use:
      // - the intended RepairStrategy (if != null)
      // - the "default" RepairStrategy (implemented at "buildDefaultRepairStrategy_FromTo")
      E_RepairStrategy_FromTo lC_repairStrategy = aC_intendedRepairStrategy;
      if( lC_repairStrategy == null )
      {
         lC_repairStrategy = buildDefaultRepairStrategy_FromTo();
      }
      ArrayList larr_repairPlan = lC_repairStrategy.eval( lcl_origin, lcl_destination );
      return larr_repairPlan;
   }


   /**
    * Default RepairStrategy:
    * - "ShorterRoute"
    */
   private E_RepairStrategy_FromTo buildDefaultRepairStrategy_FromTo()
   {
      E_RepairStrategy_FromTo lC_defaultRepairStrategy =
      new E_RepairStrategy_FromTo()
      {
         public ArrayList eval( MotionlessObject acl_origin, MotionlessObject acl_destination, Collection acl_tabooLocations )
         {
            B_GIS_Router_FromTo litf_planner = new B_GIS_Router_FromTo();
            //return litf_planner.getShorterRouteFromTo( acl_origin, acl_destination, acl_tabooLocations );
            return litf_planner.getPassableRouteFromTo( acl_origin, acl_destination, acl_tabooLocations );
         }
      };
      return lC_defaultRepairStrategy;
   }


   private void updateIndexOfCurrentLocation()
   {
      iint_indexOfCurrentLocation = findIndexOfCurrentLocation();
   }


   private void updateIndexOfExpectedLocationAtNextTime()
   {
      if( iint_indexOfCurrentLocation < 0 )
      {
         return;
      }
      iint_indexOfExpectedLocationAtNextTime = findIndexOfExpectedLocationAtNextTime();
   }


   /**
    * Return the index of current location
    * or -1 if current location not found at "iarr_planToMonitor"
    */
   private int findIndexOfCurrentLocation()
   {
      MotionlessObject lcl_agentLocation = icl_agentObjectToMonitor.motionlessPosition();
      int lint_index = 0;
      for( Iterator litf_iterator = iarr_planToMonitor.iterator();
           litf_iterator.hasNext(); )
      {
         MotionlessObject lcl_planLocation = ( MotionlessObject )litf_iterator.next();
         if( lcl_agentLocation == lcl_planLocation )
         {
            return lint_index;
         }
         ++lint_index;
      }
      return -1;
   }


   /**
    * Return the index of expected location location at next time
    * or (iarr_planToMonitor.size()-1) if "next time" is "after iarr_planToMonitor" length
    */
   private int findIndexOfExpectedLocationAtNextTime()
   {
      int lint_index = iint_indexOfCurrentLocation;
      int lint_distanceSum = 0;
      boolean lbln_progress = true;
      while( lbln_progress &&
             ( lint_index < iarr_planToMonitor.size() - 1 ) &&
             ( lint_index >= 0 ) )
      {
         MotionlessObject lcl_nextLocation = ( MotionlessObject )iarr_planToMonitor.get( lint_index + 1 );
         if( lcl_nextLocation instanceof Road )
         {
            lint_distanceSum = lint_distanceSum + ( ( Road )lcl_nextLocation ).length();
            lbln_progress = ( lint_distanceSum < pint_maximumExpectedProgressAtEachTime);
         }
         ++lint_index;
      }
      lint_index = ( lbln_progress ) ?
                   ( iarr_planToMonitor.size() - 1 )
                   :
                   ( lint_index - 1 );
      return lint_index;
   }


   public void printlnAll()
   {
      U_LeitorEscritorLinha.escrever( "" );
      U_LeitorEscritorLinha.escrever( "-----" );
      U_LeitorEscritorLinha.escreverMesmaLinha( "maximumExpectedProgressAtEachTime = " );
      U_LeitorEscritorLinha.escrever( U_Geral.paraString( pint_maximumExpectedProgressAtEachTime) );
      U_LeitorEscritorLinha.escreverMesmaLinha( "admissibleDelay = " );
      U_LeitorEscritorLinha.escrever( U_Geral.paraString( pint_admissibleDelay) );
      U_LeitorEscritorLinha.escreverMesmaLinha( "agentObjectToMonitor ID = " );
      U_LeitorEscritorLinha.escrever( U_Geral.paraString( icl_agentObjectToMonitor.id ) );
      U_LeitorEscritorLinha.escreverMesmaLinha( "timeON = " );
      U_LeitorEscritorLinha.escrever( U_Geral.paraString( iint_timeON ) );
      U_LeitorEscritorLinha.escreverMesmaLinha( "timeCurrent = " );
      U_LeitorEscritorLinha.escrever( U_Geral.paraString( iint_timeCurrent ) );
      U_LeitorEscritorLinha.escreverMesmaLinha( "PLAN TO MONITOR:: " );
      for( Iterator litf_iterator = iarr_planToMonitor.iterator(); litf_iterator.hasNext(); )
      {
         MotionlessObject lcl_object = ( MotionlessObject )litf_iterator.next();
         U_LeitorEscritorLinha.escreverMesmaLinha( U_Geral.paraString( lcl_object.id ) + " " );
      }
      U_LeitorEscritorLinha.escrever( "" );
      U_LeitorEscritorLinha.escreverMesmaLinha( "indexOfCurrentLocation = " );
      U_LeitorEscritorLinha.escrever( U_Geral.paraString( iint_indexOfCurrentLocation ) );
      U_LeitorEscritorLinha.escreverMesmaLinha( "indexOfExpectedLocationAtNextTime = " );
      U_LeitorEscritorLinha.escrever( U_Geral.paraString( iint_indexOfExpectedLocationAtNextTime ) );
      U_LeitorEscritorLinha.escreverMesmaLinha( "accumulatedDeviation = " );
      U_LeitorEscritorLinha.escrever( U_Geral.paraString( iint_accumulatedDeviation ) );
      U_LeitorEscritorLinha.escreverMesmaLinha( "needToRepair = " );
      U_LeitorEscritorLinha.escrever( U_Geral.paraString( getNeedToRepairPlan () ) );
      U_LeitorEscritorLinha.escreverMesmaLinha( "unexpectedCost = " );
      U_LeitorEscritorLinha.escrever( U_Geral.paraString( iint_unexpectedCost ) );
      U_LeitorEscritorLinha.escrever( "" );
   }


   private void println( String astr_message )
   {
      System.out.println( astr_message );
   }


   /**
    * Strategy Specification
    * Used at:
    * - "buildRepairPlan"
    * - "buildRepairPlanWhenExecutionOutOfPlan"
    */
   abstract private static class E_RepairStrategy_FromTo
   {
      abstract public ArrayList eval( MotionlessObject acl_origin, MotionlessObject acl_destination, Collection acl_tabooLocations );
      public ArrayList eval( MotionlessObject acl_origin, MotionlessObject acl_destination )
      {
         return eval( acl_origin, acl_destination, new ArrayList() );
      }
   }
}
