/* PTS */


package rCr.agent.skillLevel_1.problemSolver;


import java.util.Collection;
import java.util.HashSet;
import java.util.ArrayList;
import yab.agent.object.MotionlessObject;
import java.util.Iterator;
import yab.agent.object.Building;
import rCr.util.U_Geral;
import rCr.util.E_Condition;


/**
 *  @author PTS
 *  @since MAY.2004
 */
public class B_GIS_RoutingProblem
extends B_Problem
{
   protected final HashSet goalStateSet;
   // Taboo (Forbidden) Locations
   protected final HashSet tabooLocations;
   protected final C_Searcher solutionSearcher;


   public B_GIS_RoutingProblem( Collection litf_destinations,
                                Collection litf_tabooLocations )
   {
      // Set a Problem with a "not yet specified" Initial State
      this( null, litf_destinations, litf_tabooLocations );
   }


   public B_GIS_RoutingProblem( MotionlessObject acl_gisLocation,
                                Collection litf_destinations,
                                Collection litf_tabooLocations )
   {
      // ------------------------------------------------------
      // Set the Taboo (Forbidden) Locations
      // The state space excludes these taboo locations
      // (state specification includes an invariant to
      //  guarantee that a state never belongs to a taboo list)
      // ------------------------------------------------------
      this.tabooLocations = U_Geral.toHashSet( litf_tabooLocations );

      // ---------------------------
      // Initial State Specification
      // ---------------------------
      // if the initial state is not specified now
      // then, it must be specified before execution of "solve" method
      if( null != acl_gisLocation )
      {
         specifyRoutingProblemInitialState( acl_gisLocation );
      }

      // ------------------
      // Goal Specification
      // ------------------
      this.goalStateSet = U_Geral.toHashSet( litf_destinations );

      // --------------------------------------
      // Add each Operator (to the OperatorSet)
      // --------------------------------------
      addOperatorSet();

      // ----------------------------
      // Create the intended Strategy
      // ----------------------------
      I_SearchStrategy litf_strategy = getIntendedStrategy();

      // -----------------------------------------------
      // Create Solution Searcher with intended Strategy
      // -----------------------------------------------
      solutionSearcher = new C_Searcher( litf_strategy );
   }


   /**
    * To be redefined by subclasses with other specific conditions
    */
   protected E_Condition getStateInvariant()
   {
      // Default: a route between 2 locations should not explore Buildings
      E_Condition lC_notBuildingExceptAtGoal = getCondition_notBuildingExceptAtGoal();
      // Default "State Invariant"
      E_Condition lC_stateInvariant = lC_notBuildingExceptAtGoal;
      if( ! tabooLocations.isEmpty() )
      {
         // Taboo Locations are not visited
         E_Condition lC_notTabooLocation = getCondition_notTabooLocation();
         // Set "State Invariant" with default "State Invariant" and "Taboo Locations"
         lC_stateInvariant = lC_notBuildingExceptAtGoal.and( lC_notTabooLocation );
      }
      return lC_stateInvariant;
   }


   /**
    * The initial state specification may be done
    * - during this class construction
    * - after the class construction (and before execution of "solve" method)
    */
   public void specifyRoutingProblemInitialState( Object acl_gisLocation )
   {
      MotionlessObject lcl_gisLocation = ( MotionlessObject )acl_gisLocation;
      // ---------------------------
      // Initial State Specification
      // ---------------------------
      E_Condition lC_stateInvariant = getStateInvariant ();
      E_State lcl_initialState = new E_State( lcl_gisLocation, lC_stateInvariant );

      // Set the initial state
      setInitialState( lcl_initialState );
   }


   private E_Condition getCondition_notBuildingExceptAtGoal()
   {
      E_Condition lC_notBuildingExceptAtGoal =
      new E_Condition()
      {
         public boolean eval( Object lcl_object )
         {
            boolean lbln_eval = true;
            MotionlessObject lcl_gisLocation = ( MotionlessObject )lcl_object;
            boolean lbln_isBuilding = ( lcl_gisLocation instanceof Building );
            // Only consider Buildings if they are contained at the Goal State
            if( lbln_isBuilding )
            {
               lbln_eval = goalStateSet.contains( lcl_gisLocation );
            }
            return lbln_eval;
         }
      };
      return lC_notBuildingExceptAtGoal;
   }


   private E_Condition getCondition_notTabooLocation()
   {
      E_Condition lC_notTabooLocation =
      new E_Condition()
      {
         public boolean eval( Object lcl_object )
         {
            boolean lbln_eval = true;
            MotionlessObject lcl_gisLocation = ( MotionlessObject )lcl_object;
            // Taboo Locations are disallowed
            lbln_eval = ( ! tabooLocations.contains( lcl_gisLocation ) );
            return lbln_eval;
         }
      };
      return lC_notTabooLocation;
   }


   /**
    * To be redefined by subclasses with other operator set
    */
   protected void addOperatorSet()
   {
      addOperator( new E_GIS_move() );
   }


   /**
    * To be redefined by subclasses with other specific Strategy Configuration
    */
   protected I_SearchStrategy getIntendedStrategy()
   {
      /**
       * Search Strategy: AStar WITHOUT Repeated Nodes (in the whole search tree)
       * Inline Construction of a derivate class (from C_AStar)
       */
      I_SearchStrategy litf_strategy =
      // new C_UniformCost()
      new C_AStar()
      {
         // The set of already visited "Gis Locations"
         private HashSet icl_visitedGisLocations = new HashSet();

         public boolean preStateExpandCondition( E_Node acl_currentNode )
         {
            E_State lcl_state = ( E_State )acl_currentNode.getState();
            // Register this as an already visited "gisLocation"
            // return true: because the state must be expanded
            icl_visitedGisLocations.add( lcl_state.gisLocation );
            return true;
         }

         public boolean postStateExpandCondition( E_Node acl_parentNode, I_State aitf_aNextState )
         {
            E_State lcl_state = ( E_State )aitf_aNextState;
            boolean lbln_visitedGisLocation = icl_visitedGisLocations.contains( lcl_state.gisLocation );
            // Do not add to the fringe queue a previously visited "gisLocation"
            return ( ! lbln_visitedGisLocation );
         }
      };
      return litf_strategy;
   }


   public ArrayList solve()
   {
      if( ! isInitialStateSpecified() )
      {
         throw new Error( "PTS: B_GIS_RoutingProblem::slolve initial state not specified!!" );
      }

      return solutionSearcher.searchSolution( this );
   }


   boolean isGoalState( I_State acl_state )
   {
      E_State lcl_state = ( E_State )acl_state;
      MotionlessObject lcl_gisLocation = lcl_state.gisLocation;
      return goalStateSet.contains( lcl_gisLocation );
   }


   /**
    * O custo estimado do "estado-actual (acl_state)" at a um estado objectivo
    * h - tambm se designa por "distancia heuristica ate' ao objectivo"
    * Se a distancia for optimista - admissivel, entao A* encontra solucoes optimas
    * A distancia por omissao podera' ser zero - que certamente e' optimista
    * Necessario redefinir este mtodo caso a subclasse de problema tenha:
    * - uma ideia diferente para esta estimativa
    */
   public float getHeuristicCost( I_State acl_state )
   {
      return 0;
   }


   /**
    * State definition for this problem
    */

   public static class E_State
   implements I_State
   {
      /**
       * GIS Location
       */
      public final MotionlessObject gisLocation;
      /**
       * A problem specific condition to be
       * validated during the application of each operator.
       * This technique enables to:
       * - build higher level (more general purpose) operators
       * - add restrictions to those higher level operators
       * - enable restrictions to refer problem specific data
       * e.g. Operator: expand neighbour nodes
       *      Restriction1: do not consider Buildings when searching for a path between two locations
       *      Restriction2: never enter a specific Road (e.g. with a traffic jam)
       */
      public final E_Condition stateInvariant;


      public E_State( MotionlessObject acl_gisLocation,
                      E_Condition aC_stateInvariant )
      {
         this.gisLocation = acl_gisLocation;
         this.stateInvariant = aC_stateInvariant;
      }
   }
}
