/* PTS */


package rCr.agent.skillLevel_1.problemSolver;


import java.util.ArrayList;


/**
 *  @author PTS
 *  @since MAY.2004
 */
abstract class B_Problem
implements I_ProblemSolver
{
   private I_State icl_initialState = null;
   private ArrayList iarr_operatorSet = new ArrayList( 1 );
   // Counts the total number of expanded nodes
   private int iint_totalExpandedNodes = 0;
   // concentrationFactor is an efficiency measure of the algorithm
   // concentrationFactor = solutionDepth / totalExpandedNodes
   // if concentrationFactor = 1, then only solution nodes where expanded
   // if concentrationFactor "close to" 0, great number of non solution expanded nodes
   private int iint_concentrationFactor = 0;


   public abstract ArrayList solve();
   abstract boolean isGoalState( I_State acl_state );
   /**
    * O custo estimado do "estado-actual (acl_state)" at a um estado objectivo
    * h - tambm se designa por "distancia heuristica ate' ao objectivo"
    * Se a distancia for optimista - admissivel, entao A* encontra solucoes optimas
    * A distancia por omissao podera' ser zero - que certamente e' optimista
    * Necessario redefinir este mtodo caso a subclasse de problema tenha:
    * - uma ideia diferente para esta estimativa
    */
   public abstract float getHeuristicCost( I_State acl_state );


   protected void addOperator( I_Operator aitf_operator )
   {
      iarr_operatorSet.add( aitf_operator );
   }


   protected void setInitialState( I_State acl_initialState )
   {
      this.icl_initialState = acl_initialState;
   }


   protected boolean isInitialStateSpecified()
   {
      return ( icl_initialState != null );
   }


   public I_State getInitialState()
   {
      return icl_initialState;
   }


   public ArrayList getOperatorSet()
   {
      return iarr_operatorSet;
   }


   public int getTotalExpandedNodes()
   {
      return this.iint_totalExpandedNodes;
   }


   void setTotalExpandedNodes( int aint_totalExpandedNodes )
   {
      this.iint_totalExpandedNodes = aint_totalExpandedNodes;
   }


   public int getConcentrationFactor()
   {
      return this.iint_concentrationFactor;
   }


   void setConcentrationFactor( int aint_concentrationFactor )
   {
      this.iint_concentrationFactor = aint_concentrationFactor;
   }
}
