/* PTS */


package rCr.agent.skillLevel_1.problemSolver;


import java.util.ArrayList;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.Iterator;
import rCr.util.E_Condition;


class C_Searcher
{
   private I_SearchStrategy searchStrategy;
   private ArrayList solution;
   private LinkedList fringeNodesQueue;
   private HashSet closedStates;


   public C_Searcher( I_SearchStrategy aitf_searchStrategy )
   {
      this.searchStrategy = aitf_searchStrategy;
   }


   public ArrayList searchSolution( B_Problem acl_problem )
   {
      solution = new ArrayList();
      fringeNodesQueue = new LinkedList();
      closedStates = new HashSet();

      // Create initial node (fringe) from "initial state"
      E_Node lcl_node = new E_Node ( acl_problem.getInitialState() );
      fringeNodesQueue.addFirst( lcl_node );

      boolean lbln_solutionFound = false;
      I_State lcl_state;
      while ( ! lbln_solutionFound &&
              ! ( fringeNodesQueue.isEmpty() ) )
      {
         // System.out.println( "Queue size is " + fringeNodesQueue.size() );
         lcl_node = ( E_Node )fringeNodesQueue.removeFirst();
         lcl_state = lcl_node.getState();

         if( acl_problem.isGoalState( lcl_state ) )
         {
            solution = lcl_node.getPathFromRoot();
            // Set the "concentration factor" value of the algorithm
            // concentrationFactor = solutionDepth / totalExpandedNodes
            int lint_totalExpandedNodes = acl_problem.getTotalExpandedNodes();
            int lint_concentrationFactor = ( lint_totalExpandedNodes <= 0 )?
                                           -1
                                           :
                                           lcl_node.getDepth() / lint_totalExpandedNodes;
            acl_problem.setConcentrationFactor( lint_concentrationFactor );
            lbln_solutionFound = true;
         }
         else
         {
            ArrayList larr_reachableNodes = expand( acl_problem, lcl_node );
            if( ! larr_reachableNodes.isEmpty() )
            {
               searchStrategy.addNodesToQueue( larr_reachableNodes, fringeNodesQueue );
            }
         }
      }
      return solution;
   }


   private ArrayList expand( B_Problem acl_problem, E_Node acl_currentNode )
   {
      I_State lcl_currentState = acl_currentNode.getState();
      /**
       * Exemplo utilizacao da "conditionPreExpand":
       * nao expandir um elemento para
       * alem de determinada profundidade
       */
      boolean lbln_preStateExpandCondition = searchStrategy.preStateExpandCondition( acl_currentNode  );
      ArrayList larr_resultNodeSet = new ArrayList( 4 );
      if( lbln_preStateExpandCondition )
      {
         // Aply Each Operator (specified for this problem)
         // -----------------------------------------------
         for( Iterator litf_operatorSet = acl_problem.getOperatorSet().iterator();
              litf_operatorSet.hasNext(); )
         {
            // Get Operator
            I_Operator litf_operator = ( I_Operator )litf_operatorSet.next();

            // Apply Operator to build next State Set
            ArrayList larr_reachableStateSet = litf_operator.getNextStateSet( lcl_currentState );

            // Filter States (reachable) using Strategy "postStateExpandCondition"
           /**
            * Exemplo utilizacao da "postStateExpandCondition":
            * nao considerar um elemento cujo
            * estado ja tenha sido considerado
            * Tres situacoes:
            * - como pai
            * - no mesmo ramo
            * - em toda a arvore (implementado)
            */
            ArrayList larr_reachableStateSetFiltered = filter( acl_currentNode, larr_reachableStateSet );

            // Create a Node for each State
            larr_resultNodeSet = createNodeSetFromStateSet( acl_problem,
                                                            acl_currentNode,
                                                            larr_reachableStateSetFiltered,
                                                            litf_operator );

            // Updates the total number of expanded nodes
            int lint_expandedNodes = larr_resultNodeSet.size();
            acl_problem.setTotalExpandedNodes( lint_expandedNodes +
                                               acl_problem.getTotalExpandedNodes() );
         }
      }
      return larr_resultNodeSet;
   }


   private ArrayList createNodeSetFromStateSet( B_Problem acl_problem,
                                                E_Node acl_parentNode,
                                                ArrayList larr_reachableStateSet,
                                                I_Operator litf_operator )
   {
      I_State lcl_parentState = acl_parentNode.getState();
      ArrayList larr_resultNodeSet = new ArrayList( 4 );
      // Iterate the reachable State Set
      // -------------------------------
      for( Iterator litf_stateSet = larr_reachableStateSet.iterator();
           litf_stateSet.hasNext(); )
      {
         I_State lcl_state = ( I_State )litf_stateSet.next();
         float lflt_stateExpandCost = litf_operator.getCost( lcl_parentState, lcl_state );
         float lflt_gCost = acl_parentNode.g_cost + lflt_stateExpandCost;
         float lflt_hCost = acl_problem.getHeuristicCost( lcl_state );
         /**
          * =========================================
          * [Russell & Norvig], pag. 99 AIMA 2nd Ed.:
          * =========================================
          * In order not to expand repeated nodes (GraphSearch) and
          * still have an optimal solution using A*
          * it is necessary to ensure that the optimal path
          * to any repeated state is always the first one followed.
          * This property holds if the requirement of
          * consistency (also called monotonicity) on h_cost:
          * h(n) <= c(n,a,n') + h(n')
          * <=>
          * h(n') >= h(n) - c(n,a,n')
          * n' is: the sucessor of n generated by action a
          * To ensure heuristic consistency
          * hCost ("lflt_hCost") must be corrected
          * ---------------
          * IMPORTANT NOTE:
          * ---------------
          * If we do not want to always apply this correction then:
          * - make an abstract method in this class (e.g. applyHeuristicConsistency)
          * - make a subclass (e.g. GraphSearch) and
          *   implement that method with the hCost correction (implemented bellow)
          * - make another subclass (e.g. TreeSearch)
          *   that return the received hCost (identity function)
          * - delete hCost correction from the next lines of code (implemented bellow)
          */
         // Apply correction to guarantee "Heuristic Consistency"
         lflt_hCost = ( lflt_hCost >= acl_parentNode.h_cost - lflt_stateExpandCost ) ?
                      lflt_hCost
                      :
                      acl_parentNode.h_cost - lflt_stateExpandCost;

         // f = g + h
         float lflt_fCost = lflt_gCost + lflt_hCost;

         // Create a New Node
         E_Node lcl_node = new E_Node( lcl_state,
                                       acl_parentNode,
                                       lflt_gCost,
                                       lflt_hCost,
                                       lflt_fCost,
                                       litf_operator.getAction() );

         // Add Node to Result Node Set
         larr_resultNodeSet.add( lcl_node );
      }
      return larr_resultNodeSet;
   }


   private ArrayList filter( final E_Node acl_parentNode, ArrayList aarr_reachableStateSet )
   {
      /**
       * Para cada elemento resultante da expansao
       * vai ser verificada a pos condicao
       * Apenas os elementos que verificarem
       * a pos condicao pertencem 'a lista resultante
       */
      // Set the "postStateExpandCondition"
      E_Condition lC_postStateExpansion =
      new E_Condition()
      {
         public boolean eval( Object lcl_object )
         {
            return searchStrategy.postStateExpandCondition( acl_parentNode, ( I_State )lcl_object );
         }
      };

      /**
       * Extract:
       * - from the Collection "aarr_reachableStateSet"
       * - the elements that satisfy the filter condition ("postStateExpandCondition")
       * and return
       */
      return lC_postStateExpansion.extract( aarr_reachableStateSet );
   }


   public ArrayList getSolution()
   {
      return solution;
   }
}
