/* PTS */

package rCr.agentModel;


import rCr.initiator.I_InitiatorCommunicator;
import rCr.util.U_Geral;
import rCr.util.U_Depurador;
import rCr.util.U_LeitorEscritorLinha;
import rCr.communicativeAct.*;
import yab.io.KaConnectOk;
import yab.agent.DisasterSpace;
import yab.agent.object.RealObject;
import rCr.util.I_Depurador;
import yab.io.KaSense;
import yab.io.KaHear;
import rCr.initiator.I_Config;
import rCr.util.X_RcR_Accao;
import yab.agent.object.Property;
import rCr.agent.I_Agent;
import rCr.agent.skillLevel_0.B_PrimaryCapability;
import rCr.agent.skillLevel_0.I_PrimaryCapability_AgentModel;
import yab.io.ProtocolConstants;


/*
EXEMPLO da execucao de uma Thread
partindo de uma classe que implementa "Runnable"
class PrimeRun
implements Runnable
{
   long minPrime;
   PrimeRun( long minPrime )
   { this.minPrime = minPrime; }

   public void run()
   { // compute primes larger than minPrime ...
   }
}
 
   // The following code would then create a thread and start it running:
   PrimeRun p = new PrimeRun( 143 );
   new Thread( p ).start();
*/


/**
 *  @author PTS
 *  @since MAY.2004
 */
abstract public class C_RcR
implements I_AgentModel
{
   private C_Sensor icl_sensor;


   private I_PrimaryCapability_AgentModel primaryCapability;
   private int iint_hearingLimit;
   //private int agentType = 43;

   // Ambiente (Mundo) onde os Agentes estao imersos
   //protected DisasterSpace icl_disasterSpace;

   // Numero de actos de discurso recebidos e enviados
   // entre cada ciclo (entre dois "KaSense")
   private int iint_actosDiscursoRecebidos = 0;

   /**
    * To be specified by each AgentModel
    */
   // Pre Simulation Cycle Processing
   abstract protected void preSimulation();
   // World Representation
   abstract protected void updateWorld( Object lobj_percepcao );
   // Action Decision
   abstract protected void preAct();
   abstract protected void act() throws X_RcR_Accao;
   abstract protected void postAct();
   // Communication Decision
   abstract protected boolean agentHearsItself( int aint_senderID );
   abstract protected boolean acceptHearSay( int aint_senderID, int aint_remainingHearingUtterances );
   abstract protected boolean acceptHearTell( int aint_senderID, int aint_remainingHearingUtterances );
   abstract protected void handleHearSay( int aint_senderID, String astr_message );
   abstract protected void handleHearTell( int aint_senderID, String astr_message );


   public void setSensor( C_Sensor acl_sensor )
   {
      this.icl_sensor = acl_sensor;
   }


   /**
    * Some capabilities have simulation resctrictions
    * say & tell: have bounds within a simulation cycle
    */
   public void setPrimaryCapability( B_PrimaryCapability acl_primaryCapability )
   {
      this.primaryCapability = ( I_PrimaryCapability_AgentModel )acl_primaryCapability;
   }


   /**
    * A RoboCupRescue restriction
    * - specified within the rules of each competition
    */
   public void setHearingLimit( int aint_hearingLimit )
   {
      this.iint_hearingLimit = aint_hearingLimit;
   }


   public void run()
   {
      // Process before simualtion cycle begins
      preSimulation();

      final int kint_kaSense = 0;
      final int kint_kaHear = 1;
      final int kint_ERROR = -1;
      int lint_tipoPercepcao = kint_ERROR;

      U_Depurador.escrever( I_Depurador.kint_depurarInfo,
                            U_Geral.paraNomeClasse( this ),
                            "----- run -----" );

      // ----------------------------------------------
      // Ciclo de tratamento do comportamento do Agente
      // ----------------------------------------------
      int lint_numberOfSimulationCycles = 0;
      while( true )
      {
         // Recepcao da percepcao enviada pelo Kernel
         // -----------------------------------------
         Object lobj_percepcao = icl_sensor.obterPercepcao();

         // Verificacao do tipo de percepcao recebida
         // -----------------------------------------
         lint_tipoPercepcao = ( icl_sensor.ePercepcaoKaSense( lobj_percepcao ) ) ?
                              kint_kaSense
                              :
                              ( icl_sensor.ePercepcaoKaHear( lobj_percepcao ) ) ?
                              kint_kaHear
                              :
                              kint_ERROR;

         U_Depurador.escrever( I_Depurador.kint_depurarVerboso,
                               U_Geral.paraNomeClasse( this ),
                               Thread.currentThread().getName() );
         // ---------------------------------------------
         // Tratamento de cada tipo de percepcao
         // (KaSense marca fim de um "tempo de simulacao)
         // ---------------------------------------------
         /**
          * For each message sent to an agent,
          * this agent will receive one KaHear message from the kernel.
          * The KaSense is the limit of a turn,
          * it is why the counter is reset to zero.
          * Between two KaSense, an agent can receive many KaHear
          * and it has to choose which one to listen to,
          * in order to respect the hearing limit.
          * Sbastien Paquet - Leader of PAKRescueTeam
          * http://www.damas.ift.ulaval.ca/projets/RobocupRescue/index.php
         */
         switch( lint_tipoPercepcao )
         {
            case kint_kaSense:
               // Inform the agent about this new "world perception"
               updateWorld( lobj_percepcao );

               // Reinicia o numero de actos de discurso recebidos e enviados
               // -----------------------------------------------------------
               iint_actosDiscursoRecebidos = 0;
               // Inform the Agent "Primary Capability"
               // that a simulation cycle is about to begin
               // -----------------------------------------
               lint_numberOfSimulationCycles++;
               primaryCapability.beginSimulationCycle(); //iint_actosDiscursoEnviados = 0;

               // -----------
               // Actividade:
               // - antes de actuar
               // - actuar: as accoes devolvem excepcao X_RcR_Accao
               // - depois de actuar
               // ------------------
               preAct();
               try
               {
                  act();
               }
               catch( X_RcR_Accao lexc_rCrAccao )
               {
                  postAct();
               }
               catch( Error lexc_error )
               {
                  U_LeitorEscritorLinha.escrever( "rCr.agentModel.C_RcR::run catch Error exception" );
                  U_LeitorEscritorLinha.escrever( lexc_error );
               }
               catch( Exception lexc_exception )
               {
                  U_LeitorEscritorLinha.escrever( "rCr.agentModel.C_RcR::run catch Exception exception" );
                  U_LeitorEscritorLinha.escrever( lexc_exception );
               }
            break;

            case kint_kaHear:
               KaHear lobj_kaHear = ( KaHear )lobj_percepcao;
               /**
                * Only handle a message if:
                * - it was not previously sent by itself!, and
                * - it is still possible to hear the message (hearing limit)
                */
               if( ! agentHearsItself( lobj_kaHear.senderId )
                   &&
                   ( iint_actosDiscursoRecebidos < iint_hearingLimit ) )
               {
                  /**
                   * PTS, 25.FEV.2004
                   * KA_HEAR_SAY e KA_HEAR_TELL surgiram depois da versao .43 do RoboCupRescue
                   * KA_HEAR foi descontinuado
                   * KA_HEAR_SAY = natural voice utterance
                   * KA_HEAR_TELL = telecomunication utterance
                   */
                  // Reamaining hearing utterances (if this one is accepted)
                  int lint_remainingHearingUtterances = iint_hearingLimit - iint_actosDiscursoRecebidos - 1;
                  switch( lobj_kaHear.type )
                  {
                     case ProtocolConstants.KA_HEAR_SAY:
                        if( acceptHearSay( lobj_kaHear.senderId, lint_remainingHearingUtterances ) )
                        {
	                        iint_actosDiscursoRecebidos++;
                           handleHearSay( lobj_kaHear.senderId, lobj_kaHear.message );
                        }
                     break;
                     case ProtocolConstants.KA_HEAR_TELL:
                        if( acceptHearTell( lobj_kaHear.senderId, lint_remainingHearingUtterances ) )
                        {
	                        iint_actosDiscursoRecebidos++;
                           handleHearTell( lobj_kaHear.senderId, lobj_kaHear.message );
                        }
                     break;
                     default:
                        // Deprecated "lobj_kaHear.type"
                     break;
                  }
               }
            break;

            case kint_ERROR:
               throw new Error( ">>> ERRO ERRO: Recepcao de percepcao nao reconhecida <<<" );

            default:
               throw new Error( ">>> ERRO ERRO: !!!SITUACAO NAO PREVISTA!!! <<<" );
         }
      }
   }
}