/* PTS & AR */


package rCr.agentWorkingMemory;


import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;
import rCr.util.E_Condition;
import rCr.util.U_LeitorEscritorLinha;
import yab.agent.object.MotionlessObject;
import yab.agent.DisasterSpace;
import yab.agent.object.Node;


/**
 * @author aremedios
 * @review PTS
 * @since JUN.2004
 */
public class E_SpaceGrid
{
   /**
    * The total space section
    * (the one covered by a grid of sections)
    */
   private E_SectionWithRoads icl_fullSpaceSection = null;
   /**
    * @link aggregation
    * @associates rCr.agentWorkingMemory.E_SectionWithRoads
    * @supplierCardinality 0..*
    */
   private List ilst_sectionList = new ArrayList();
   // Number of Columns and Lines of Grid
   private int nColumns;
   private int nLines;


   public E_SpaceGrid( int aint_columns, int aint_lines, DisasterSpace acl_disasterSpace )
   {
      nColumns = aint_columns;
      nLines = aint_lines;
      ilst_sectionList = new ArrayList( nColumns * nLines );

      // Build each Space Section:
      // - "fullSpaceSection" and
      // - each Section at "ilst_sectionList"
      buildEachSpaceSection( acl_disasterSpace );
   }


   public E_SectionWithRoads getHOME( int aint_agentAge )
   {
      E_SectionWithRoads lcl_HOME = getSectionByID( aint_agentAge );
      // If empty HOME
      // then return fullSpaceSection
      if( lcl_HOME.getHOME() == null )
      {
         return icl_fullSpaceSection;
      }
      return lcl_HOME;
   }


   /**
    * id: 1, 2, 3, ... ==>> grid element at id-1 position
    * id: <= 0 or > sectionList.size() ==>> fullSpaceSection
    */
   private E_SectionWithRoads getSectionByID( int aint_id )
   {
      if( ( aint_id <= 0 ) || ( aint_id > ilst_sectionList.size() ) )
      {
         return icl_fullSpaceSection;
      }
      return ( E_SectionWithRoads )ilst_sectionList.get( aint_id - 1 );
   }


   private void setSpaceSection( E_SectionWithRoads acl_sectionWithRoads )
   {
      // The "full" SpaceSection has sectionID == 0
      acl_sectionWithRoads.setSectionID( 0 );
      this.icl_fullSpaceSection = acl_sectionWithRoads;
   }


   private void add( E_SectionWithRoads acl_sectionWithRoads )
   {
      // Set the sectionID and add to list of Sections
      int lint_sectionID = ilst_sectionList.size() + 1;
      acl_sectionWithRoads.setSectionID( lint_sectionID );
      ilst_sectionList.add( acl_sectionWithRoads );
   }


   private void buildEachSpaceSection( DisasterSpace acl_disasterSpace )
   {
	   //E_SpaceGrid resultSpaceGrid = new E_SpaceGrid ( nColumns, nLines );

      // Set the SpaceGrid boundary
	   E_Section fullSpaceSection = buildFullSpaceSection( acl_disasterSpace );
	   int widthTotal = fullSpaceSection.getWidth();
	   int heightTotal = fullSpaceSection.getHeight();
      // println( "widthTotal = " + widthTotal );
      // println( "heightTotal = " + heightTotal );

      // 1st: build the "full" space grid (a SectionWithRoads)
      // -----------------------------------------------------
      E_SectionWithRoads lcl_newSectionWithRoads = new E_SectionWithRoads( fullSpaceSection, acl_disasterSpace );
      // Set "full" Section to "fullSpaceSection"
      setSpaceSection( lcl_newSectionWithRoads );

      // 2nd: build each section of the SpaceGrid
      // ----------------------------------------
      int sectionWidth = widthTotal / nColumns;
      int sectionHeight = heightTotal / nLines;
      for( int line = 0; line < nLines ; line ++ )
      {
         for( int column = 0; column < nColumns ; column ++ )
         {
            E_Section sectionToBuild = new E_Section();
            sectionToBuild.setTopLeft_X( fullSpaceSection.getTopLeft_X() + ( sectionWidth * column ) );
            sectionToBuild.setTopLeft_Y( fullSpaceSection.getTopLeft_Y() - ( sectionHeight * line ) );

            sectionToBuild.setBottomRight_X( sectionToBuild.getTopLeft_X() + sectionWidth );
            sectionToBuild.setBottomRight_Y( sectionToBuild.getTopLeft_Y() - sectionHeight );

            // Build SectionWithRoads
            // ----------------------
            lcl_newSectionWithRoads = new E_SectionWithRoads( sectionToBuild, acl_disasterSpace );
            // Add Section to "sectionList"
            add( lcl_newSectionWithRoads );
         }
      }
	}
 

   private E_Section buildFullSpaceSection( DisasterSpace acl_disasterSpace )
   {
	   Iterator litf_iterator = acl_disasterSpace.nodes.iterator();
	   if( ! litf_iterator.hasNext() )
      {
         return null;
      }
      // Initial max and min values
      Node lcl_node =( Node )litf_iterator.next();
      int max_X = lcl_node.x();
      int min_X = lcl_node.x();
      int max_Y = lcl_node.y();
      int min_Y = lcl_node.y();

      for ( ; litf_iterator.hasNext(); lcl_node = ( Node )litf_iterator.next() )
	   {
	      if( lcl_node.x() > max_X )
	         max_X = lcl_node.x();
	      if( lcl_node.x() < min_X )
	         min_X = lcl_node.x();

         if( lcl_node.y() > max_Y )
	         max_Y = lcl_node.y();
	      if( lcl_node.y() < min_Y )
	         min_Y = lcl_node.y();
	   }
      // Buil World Section
      E_Section worldSection = new E_Section();
	   worldSection.setTopLeft_X( min_X );
	   worldSection.setTopLeft_Y( max_Y );
	   worldSection.setBottomRight_X( max_X );
	   worldSection.setBottomRight_Y( min_Y );
      return worldSection;
   }


   public void printlnAll()
   {
      U_LeitorEscritorLinha.escrever( "-----" );
      U_LeitorEscritorLinha.escrever( "SpaceGrid( nColumns, nLines ) = " +
         "( " + nColumns + ", " + nLines + " )" );
      icl_fullSpaceSection.printlnAll();
      for( Iterator litf_iterator = ilst_sectionList.iterator(); litf_iterator.hasNext(); )
      {
         U_LeitorEscritorLinha.escrever( "" );
         U_LeitorEscritorLinha.escrever( "" );
         E_SectionWithRoads lcl_sectionWithRoads = ( E_SectionWithRoads )litf_iterator.next();
         lcl_sectionWithRoads.printlnAll();
      }
   }


   private static void println( String astr_message )
   {
      System.out.println( astr_message );
   }
}
