/* PTS */


package rCr.agentWorkingMemory;


/**
 * @author PTS
 * @since JUN.2004
 * @dependency E_Point (the only dependency)
 * @review ALR
 */
public class U_Geometry
{
   /**
    * @author PTS
    * @since JUN.2004
    * ----
    * p0==(x0,y0)
    * p1==(x1,y1)
    * ----
    * d=sqrt( (x0-x1)^2 + (y0-y1)^2 )
    */
   public static double getPointToPointDistance( E_Point p0, E_Point p1 )
   {
      double dx = p0.getX() - p1.getX();
      double dy = p0.getY() - p1.getY();
      return Math.sqrt( dx * dx + dy * dy );
   }


   /**
    * Utility method: to avoid E_Point object instatiation
    * (just a performance issue)
    */
   public static double getPointToPointDistance( int x0, int y0, int x1, int y1 )
   {
      double dx = x0 - x1;
      double dy = y0 - y1;
      return Math.sqrt( dx * dx + dy * dy );
   }


   /**
    * @author PTS
    * @since JUN.2004
    * Distance from p0 to the line defined by p1 and p2
    * ----
    * p0==(x0,y0)
    * Line:
    * p1==(x1,y1)
    * p1==(x2,y2)
    * ----
    * d= ( | (x2-x1)*(y1-y0) - (x1-x0)*(y2-y1) | ) / sqrt( (x2-x1)^2 + (y2-y1)^2) )
    * ----
    * @see demonstration at: www.???.??? (perguntar Ana Luisa)
    */
   public static double getPointToLineDistance( E_Point p0, E_Point p1, E_Point p2 )
   {
      int x0 = p0.getX();
      int y0 = p0.getY();
      int x1 = p1.getX();
      int y1 = p1.getY();
      int x2 = p2.getX();
      int y2 = p2.getY();

      // (x2-x1) and (x2-x1)^2
      double x2_x1 = x2 - x1;
      double x2_x1_2 = x2_x1 * x2_x1;
      // (y2-y1) and (y2-y1)^2
      double y2_y1 = y2 - y1;
      double y2_y1_2 = y2_y1 * y2_y1;
      // numerator
      double numerator = Math.abs( ( x2_x1*(y1-y0 ) - (x1-x0)*y2_y1 ) );
      double denominator = Math.sqrt( x2_x1_2 + y2_y1_2 );
      return ( numerator / denominator );
   }


   /**
    * @author PTS
    * @since JUN.2004
    * @return E_Point intersection of two lines (given by four E_Point)
    * @return null if the lines do not intersect
    * Line Equation:
    * (x - x0) / (x1 - x0) = (y - y0) / (y1 - y0)
    * <=>
    * y = lineDisplacement + lineSlope * x
    * ----
    * Intersection Point of lines R and S:
    * lineDisplacement_R + lineSlope_R * x = lineDisplacement_S + lineSlope_S * x
    * <=>
    * x = (lineDisplacement_S - lineDisplacement_R) / (lineSlope_R - lineSlope_S)
    * infinite: parallel lines (return null) if (lineSlope_R - lineSlope_S) == 0
    * infinite: parallel vertical lines (return null) (both lineSlope_R&&S == 0)
    */
   public static E_Point getLineIntersection( E_Point R_p0, E_Point R_p1,
                                              E_Point S_p0, E_Point S_p1 )
   {
      Double slope_R = lineSlope( R_p0, R_p1 );
      Double slope_S = lineSlope( S_p0, S_p1);
      // Both vertical
      if( ( slope_R == null ) && ( slope_S == null ) )
      {
         return null;
      }
      Double intersectX = null;
      Double intersectY = null;
      Double displacement_R = lineDisplacement( R_p0, R_p1, slope_R );
      Double displacement_S = lineDisplacement( S_p0, S_p1, slope_S );
      // None vertical
      if( ( slope_R != null ) && ( slope_S != null ) )
      {
         double slope_R_S = slope_R.doubleValue() - slope_S.doubleValue();
         // infinite: parallel lines
         // (return null) if (lineSlope_R - lineSlope_S) == 0
         if( slope_R_S == 0 )
         {
            return null;
         }
         double displacement_S_R = displacement_S.doubleValue() - displacement_R.doubleValue();
         intersectX = new Double( displacement_S_R / slope_R_S );
         intersectY = new Double( displacement_S.doubleValue() + slope_S.doubleValue() * intersectX.doubleValue() );
      }
      // Line_R vertical (x = x0)
      if( slope_R == null )
      {
         intersectX = new Double( R_p0.getX() );
         intersectY = new Double( displacement_S.doubleValue() + slope_S.doubleValue() * intersectX.doubleValue() );
      }
      // Line_S vertical (x = x0)
      if( slope_S == null )
      {
         intersectX = new Double( S_p0.getX() );
         intersectY = new Double( displacement_R.doubleValue() + slope_R.doubleValue() * intersectX.doubleValue() );
      }
      if( ( intersectX == null ) || ( intersectY == null ) )
      {
         throw new Error( "PTS: - intersectLines - intersectX&&Y != null expected" );
      }
      // Math.floor
      // Returns the largest (closest to positive infinity) double value
      // that is not greater than the argument and is equal to a mathematical integer
      return new E_Point( ( int )Math.floor( intersectX.doubleValue() ),
                          ( int )Math.floor( intersectY.doubleValue() ) );
   }


   /**
    * @author PTS
    * @since JUN.2004
    * @return E_Point intersection of two segment lines R and S (given by four E_Point)
    * @return null if the segment lines do not intersect
    * ----
    * R==[p0,p1]
    * S==[p0,p1]
    * ----
    * 1st get the "intersection point" of R and S
    * 2nd if "intersection point" is contained at both R and S segment bounds (envelope)
    *     then return "intersection point"
    */
   public static E_Point getSegmentLineIntersection( E_Point acl_segmentR_p0, E_Point acl_segmentR_p1,
                                                     E_Point acl_segmentS_p0, E_Point acl_segmentS_p1,
                                                     int aint_tolerance )
   {
      E_Point lcl_intersectionPoint = getLineIntersection ( acl_segmentR_p0, acl_segmentR_p1,
                                                            acl_segmentS_p0, acl_segmentS_p1);
      // See if segments are parallel (do not intersect)
      if( lcl_intersectionPoint == null )
      {
         return null;
      }
      // If intersection point is contained at both segment envelops (augmented by tolerance)
      // then the segments intersect
      boolean lbln_pointAtSegment_R = isPointAtSegmentLineEnvelope( lcl_intersectionPoint, acl_segmentR_p0, acl_segmentR_p1, aint_tolerance );
      boolean lbln_pointAtSegment_S = isPointAtSegmentLineEnvelope( lcl_intersectionPoint, acl_segmentS_p0, acl_segmentS_p1, aint_tolerance );
      boolean lbln_segmentsIntersect = lbln_pointAtSegment_R && lbln_pointAtSegment_S;

      // If segments interset, return the intersection point
      // else return null
      return ( lbln_segmentsIntersect ) ? lcl_intersectionPoint : null;
   }


   /**
    * @author PTS
    * @since JUN.2004
    * @return E_Point intersection of two segment lines R and S (given by four E_Point)
    * @return null if the segment lines do not intersect
    * ----
    * R==[p0,p1]
    * S==[p0,p1]
    * ----
    * 1st get the "intersection point" of R and S
    * 2nd if "intersection point" is contained at both R and S segment bounds (envelope)
    *     then return true
    */
   public static boolean isSegmentLineIntersection( E_Point acl_segmentR_p0, E_Point acl_segmentR_p1,
                                                    E_Point acl_segmentS_p0, E_Point acl_segmentS_p1,
                                                    int aint_tolerance )
   {
      E_Point lcl_intersectionPoint = getLineIntersection ( acl_segmentR_p0, acl_segmentR_p1,
                                                            acl_segmentS_p0, acl_segmentS_p1);
      // See if segments are parallel (do not intersect)
      if( lcl_intersectionPoint == null )
      {
         return false;
      }
      // If intersection point is contained at both segment envelops (augmented by tolerance)
      // then segments intersect
      boolean lbln_pointAtSegment_R = isPointAtSegmentLineEnvelope ( lcl_intersectionPoint, acl_segmentR_p0, acl_segmentR_p1, aint_tolerance );
      boolean lbln_pointAtSegment_S = isPointAtSegmentLineEnvelope ( lcl_intersectionPoint, acl_segmentS_p0, acl_segmentS_p1, aint_tolerance );
      boolean lbln_segmentsIntersect = lbln_pointAtSegment_R && lbln_pointAtSegment_S;
      return lbln_segmentsIntersect;
   }


   /**
    * @author PTS
    * @since JUN.2004
    * true
    * ----
    * point p is within the envelope (bounds) delimited by
    * segment point limits: p0 and p1
    * ----
    * p==(x,y)
    * p0==(x0,y0)
    * p1==(x1,y1)
    * ----
    */
   public static boolean isPointAtSegmentLineEnvelope(
      E_Point acl_point_p, E_Point acl_segment_p0, E_Point acl_segment_p1, int aint_tolerance )
   {
      int x0 = acl_segment_p0.getX();
      int y0 = acl_segment_p0.getY();
      int x1 = acl_segment_p1.getX();
      int y1 = acl_segment_p1.getY();

      // Envelope (that envolves the segment)
      int max_X = ( x0 > x1 ) ? x0 : x1;
      int max_Y = ( y0 > y1 ) ? y0 : y1;
      int min_X = ( x0 < x1 ) ? x0 : x1;
      int min_Y = ( x0 < x1 ) ? x0 : x1;

      // Set tolerance (augment the above envelope)
      max_X = ( max_X >= 0 ) ? max_X + aint_tolerance : max_X - aint_tolerance;
      max_Y = ( max_Y >= 0 ) ? max_Y + aint_tolerance : max_Y - aint_tolerance;
      min_X = ( min_X >= 0 ) ? min_X - aint_tolerance : min_X + aint_tolerance;
      min_Y = ( min_Y >= 0 ) ? min_Y - aint_tolerance : min_Y + aint_tolerance;

      // See if point is contained at the envelope (augmented by tolerance)
      int x = acl_point_p.getX();
      int y = acl_point_p.getY();
      boolean lbln_pointAtSegmentEnvelope =
         ( x <= max_X ) && ( x >= min_X ) &&
         ( y <= max_Y ) && ( y >= min_Y );
      return lbln_pointAtSegmentEnvelope;
   }


   /**
    * @author PTS
    * @since JUN.2004
    * Line Equation
    * ----
    * p0==(x0,y0)
    * p1==(x1,y1)
    * ----
    * (x - x0) / (x1 - x0) = (y - y0) / (y1 - y0)
    * <=>
    * y = lineDisplacement + lineSlope * x
    * Line Slope: (y1 - y0) / (x1 - x0)
    * Line Displacement: y0 - x0*lineSlope
    * infinite: (return null) if (x1 - x0) == 0
    */
   private static Double lineSlope( E_Point p0, E_Point p1 )
   {
      // Line: p0==(x0,y0) and p1==(x1,y1)
      // ---------------------------------
      int x0 = p0.getX();
      int y0 = p0.getY();
      int x1 = p1.getX();
      int y1 = p1.getY();
      // infinite (return null), if (x1 - x0) == 0
      // -----------------------------------------
      double line_x1_x0 = x1 - x0;
      if( line_x1_x0 == 0 )
      {
         return null;
      }
      // Line Slope:
      // (y1 - y0) / (x1 - x0)
      // ---------------------
      double slope = ( y1 - y0 ) / line_x1_x0;
      return new Double( slope );
   }


   /**
    * Simplified usage of "lineDisplacement"
    * when "lineSlope" has not yet been calculated
    */

   private static Double lineDisplacement( E_Point p0, E_Point p1 )
   {
      return lineDisplacement( p0, p1, lineSlope( p0, p1 ) );
   }

   /**
    * @author PTS
    * @since JUN.2004
    * Line Equation
    * ----
    * p0==(x0,y0)
    * p1==(x1,y1)
    * ----
    * (x - x0) / (x1 - x0) = (y - y0) / (y1 - y0)
    * <=>
    * y = lineDisplacement + lineSlope * x
    * Line Slope: (y1 - y0) / (x1 - x0)
    * Line Displacement: y0 - x0*lineSlope
    * infinite: (return null) if (x1 - x0) == 0
    */
   private static Double lineDisplacement( E_Point p0, E_Point p1, Double aDbl_slope )
   {
      // Line: p0==(x0,y0) and p1==(x1,y1)
      // ---------------------------------
      int x0 = p0.getX();
      int y0 = p0.getY();
      int x1 = p1.getX();
      int y1 = p1.getY();
      // y0 - x0*lineSlope
      // -----------------
      Double slope = aDbl_slope;
      if( slope == null )
      {
         return null;
      }
      // Line Displacement:
      // y0 - x0*lineSlope
      // -----------------
      double displacement = y0 - x0 * slope.doubleValue();
      return new Double( displacement );
   }

}