/* PTS */
/**
* Implementa o padrao Gof: Adaptaer
*/


package rCr.communicator;

import yab.io.RCRSSProtocolSocket;
import rCr.util.U_Endereco;
import rCr.util.I_Depurador;
import rCr.util.U_Depurador;
import rCr.util.U_Geral;
import yab.io.KaConnectError;
import yab.io.NozzleElement;



public class B_RCRSS
implements I_Communicator
{
   private RCRSSProtocolSocket adaptee = null;
   /**
    * @directed 
    */
   private U_Endereco lnkU_Endereco = null;


   public B_RCRSS()
   {
      super();
   }


   /**
    * Registar o endereco do no remoto (Kernel)
    * Efeito colateral:
    * - neste momento ja' e' possivel criar uma instancia do "adaptee"
    * - parametros do construtor do "adaptee":
    * - ( InetAddress kernelAddress, int kernelPort )
    */
   public void registarEnderecoProcessoRemoto( U_Endereco acl_enderecoNoRemoto )
   {
      // registar o endereco do no remoto (kernelAddress e kernelPort)
      lnkU_Endereco = acl_enderecoNoRemoto;

      // criar a instancia da classe adaptada (adaptee)
      adaptee = new RCRSSProtocolSocket( lnkU_Endereco.obterEnderecoRede(),
                                         lnkU_Endereco.obterPorto() );
   }


    public Object conectarAgente( int aint_tipoAgente )
   {
      // Enviar ao Kernel o pedido de conexao
      // (para o tipo de agente em causa)
      adaptee.akConnect( 0, aint_tipoAgente );

      U_Depurador.escrever( I_Depurador.kint_depurarVerboso,
                            U_Geral.paraNomeClasse( this ),
                            "--- Pedido de conexao enviado ao Kernel ---" );

      // Obter, a resposta do Kernel, ao pedido de conexao
      Object lobj_resposta = adaptee.receive();

      // Tratar "erro de conexao" enviado pelo Kernel
      if( lobj_resposta instanceof KaConnectError )
      {
         throw new Error( ( ( KaConnectError )lobj_resposta ).reason );
      }

      // Devolver os dados enviados pelo Kernel apos sucesso na conexao
      return lobj_resposta;
   }


   public void confirmarConexao( int aint_idAgente )
   {
      // Enviar ao Kernel confirmacao de conexao
      // (para o ID atribuido, pelo Kernel, ao agente em causa)
      adaptee.akAcknowledge( aint_idAgente );

      U_Depurador.escrever( I_Depurador.kint_depurarVerboso,
                            U_Geral.paraNomeClasse( this ),
                            "--- Confirmacao de conexao com sucesso enviado ao Kernel ---" );
   }


   public Object obterPercepcao()
   {
      // Obter percepcao, do Kernel
      return adaptee.receive();
   }


   public void difundirMensagem( String astr_mensagem, int aint_idAgente, String astr_conteudo )
   {
      // Os actos de discurso reconhecidos no RoboCupRescue
      final String kstr_akSay = "AkSay";
      final String kstr_akTell = "AkTell";

      if( astr_mensagem.equalsIgnoreCase( kstr_akSay ) )
      {
         adaptee.akSay( aint_idAgente, astr_conteudo );
         return;
      }

      if( astr_mensagem.equalsIgnoreCase( kstr_akTell ) )
      {
         adaptee.akTell( aint_idAgente, astr_conteudo );
         return;
      }
   }

   /**
    * AR, 16-FEV-2004
    * AK_UNLOAD (AmbulanceTeam);
    * AK_REST(ALL) 
    */
   public void submeterAccao( int aint_idAgente, String astr_accao )
   {
      // As accoes no RoboCupRescue
      final String kstr_akRest = "AkRest";
      final String kstr_akUnLoad = "AkUnLoad";



      if( astr_accao.equalsIgnoreCase( kstr_akRest ) )
      {
         adaptee.akRest( aint_idAgente );
         return;
      }if( astr_accao.equalsIgnoreCase( kstr_akUnLoad ) )
      {
         adaptee.akUnload( aint_idAgente );
         return;
      }

   }

   /**
    * AR, 16-FEV-2004
    * AK_EXTINGUISH (FireBrigade);
    * AK_MOVE(ALL) 
    */
   public void submeterAccao( int aint_idAgente, String astr_accao, int[] arr_argumento )
   {
      // As accoes no RoboCupRescue
      // com argumento a passar 'a accao
      final String kstr_akMove = "AkMove";

      if( astr_accao.equalsIgnoreCase( kstr_akMove ) )
      {
         adaptee.akMove( aint_idAgente, arr_argumento );
         return;
      }
   }
    public void submeterAccao( int aint_idAgente, String astr_accao, NozzleElement[] arr_argumento )
   {
      // As accoes no RoboCupRescue
      // com argumento a passar 'a accao

      final String kstr_akExtinguish = "AkExtinguish";

      if( astr_accao.equalsIgnoreCase( kstr_akExtinguish ) )
      {
         adaptee.akExtinguish( aint_idAgente, arr_argumento );
         return;
      }

   }
    /**
    * AR, 16-FEV-2004
    * AK_RESCUE, AK_LOAD (AmbulanceTeam);
    * AK_CLEAR(PoliceForce) 
    */
   public void submeterAccao( int aint_idAgente, String astr_accao, int aint_argumento )
   {
		final String kstr_akRescue = "AkRescue";
      	final String kstr_akLoad = "AkLoad";
        final String kstr_akClear = "AkClear";

 		if( astr_accao.equalsIgnoreCase( kstr_akRescue ) )
		{
        	adaptee.akRescue( aint_idAgente, aint_argumento );
         	return;
      	}
 		if( astr_accao.equalsIgnoreCase( kstr_akLoad ) )
		{
        	adaptee.akLoad( aint_idAgente, aint_argumento );
         	return;
      	}
 		if( astr_accao.equalsIgnoreCase( kstr_akClear ) )
		{
        	adaptee.akClear( aint_idAgente, aint_argumento );
         	return;
      	}
   }
}
