/* PTS */

package rCr.initiator;

import rCr.util.U_Geral;
import rCr.util.I_Depurador;
import rCr.util.U_Depurador;
import java.util.Enumeration;
import java.util.List;
import java.util.Vector;
import rCr.util.U_LeitorEscritorLinha;
import java.util.Iterator;
import rCr._BUILDER.I_BUILDER;
import rCr._BUILDER.C_BUILDER_Director;

/**
 *  @stereotype control
*/
abstract public class C_Initiator
implements I_Depurador
{
   public E_Initiator ie_initiator;
   public B_Initiator ib_initiator;
   
   /** @link aggregation
    * @directed
   * "array" de "List" com as "Threads" a executar:
   * - "array" com o numero de classes iniciadoras
   * - cada elemento do "array" e' uma "List"
   *   - cada elemento da "List" e' um objecto
   *     que sera executados numa "Thread" independente
   **/
   private List[] ilst_threads;

   /** @link dependency */
   /*#E_InitiatorCommunicator lnkE_InitiatorCommunicator;*/

   /**
    * Metodo virtual, redefinido nas subclasses
    * prepara o que for necessario antes da execucao dos Agentes
    */
   public abstract void prepararParaExecutarAgentes();


   public C_Initiator()
   {
      definirLigacoes();
   }
   

   public void iniciar( String[] astr_opcoes )
   {
      // Pre-Condicao:
      // Se nao for possivel registar as opcoes de execucao
	   // apresentar informacao de utilizacao e abandonar execucao
	   if( ! ie_initiator.registarOpcoes( astr_opcoes ) )
	   {
	      ib_initiator.utilizacao();
         return;
      }

      // Iniciar Agentes:
      // - construir respectivo objecto
      // - registar Agente na plataforma (Kernel no caso do RcR)
      // - criar a "Thread" de execucao do Agente
      construirRegistarCriarThreadParaCadaAgente ();

      // Preparar o que e' comum ao conjunto de todos as Agentes
      // (metodo virtual, redefinido nas subclasses)
      prepararParaExecutarAgentes();

      // Iniciar (start) de cada uma das "Threads"
      // (onde ira' executar cada Agente)
      iniciarThreadParaCadaAgente();
   }


   private void construirRegistarCriarThreadParaCadaAgente()
   {
      // Iniciar o "array" que ira' conter as diversas "List"
      ilst_threads = new List[ ie_initiator.obterNumeroClassesIniciadoras() ];

      // Instanciar cada uma das classes iniciadoras
      //============================================
      int lint_cicloClasse = 0;
      for( Enumeration lenm_classesIniciadoras = ie_initiator.obterClassesIniciadoras();
           lenm_classesIniciadoras.hasMoreElements();
           lint_cicloClasse++ )
      {
         // Obter cada classe iniciadora contida em "lenm_classesIniciadoras"
         Object lobj_concreteBuilderClass = lenm_classesIniciadoras.nextElement();

         // Obter o numero de threads para a classe iniciadora "lobj_classeIniciadora"
         int lint_numeroThreads = ie_initiator.obterNumeroThreads( lobj_concreteBuilderClass);

         // Iniciar o "Vector" (implementacao de "List")
         // que ira' conter a "Thread" de cada classe iniciadora
         ilst_threads[ lint_cicloClasse ] = new Vector();

         // Iniciar cada uma das "Threads" pretendidas
         //===========================================
         boolean lbln_todosAgentesJaRegistados = false;
         for( int lint_cicloThread = 0;
              ( lint_cicloThread < lint_numeroThreads )
              &&
              ( ! lbln_todosAgentesJaRegistados );
              lint_cicloThread++ )
         {
            /**
             * Builder Pattern Client:
             * - the "concrete builder" and
             * - the "builder director"
             * (GoF, pag. 98)
             */
            I_BUILDER litf_concreteBuilder = ( I_BUILDER )U_Geral.paraObject( lobj_concreteBuilderClass );
            C_BUILDER_Director lcl_builderDirector = new C_BUILDER_Director( litf_concreteBuilder );

            U_Depurador.escrever( kint_depurarVerboso,
                                  "--- ANTES: solicitar inicio Thread: " +
                                  U_Geral.paraString( lint_cicloThread ) );

            try
            {
               /**
                * Construct (using "builder director")
                */
               lcl_builderDirector.construct( new E_InitiatorCommunicator( ie_initiator ) );

               // Caso a iniciacao tenha tido sucesso:
               // - criar a Thread
               // - armazenar na "List" de "Threads"
               // - iniciar a Thread
               // ------------------------------------------------
               Thread lcl_agentThread = new Thread( litf_concreteBuilder.getResult() );
               ( ilst_threads[ lint_cicloClasse ] ).add( lcl_agentThread );
               // PTS: MAY.2004 (inicia logo a Thread)
               // (ja' nao se usa "iniciarThreadParaCadaAgente()")
               lcl_agentThread.start();
            }
            catch( Error lexc_excepcao )
            {
               String lstr_motivo = lexc_excepcao.getMessage();
               // Terminar construcao de instancias deste iniciador,
               // caso o Kernel tenha rejeitado este pedido de conexao
               // e se tenha definido como limite o numero maximo de "Threads"
               lbln_todosAgentesJaRegistados =
                  ( I_Config.kstr_motivoConexaoRecusada.equals( lstr_motivo ) )
                  &&
                  ( I_Config.kint_MAXIMO_AGENTES_ACEITE == lint_numeroThreads );
               if( lbln_todosAgentesJaRegistados )
               {
                  // Atribuir o numero de "Threads" que o Kernel aceitou
                  ie_initiator.atribuirNumeroThreads( lobj_concreteBuilderClass, lint_cicloThread );

                  U_Depurador.escrever( kint_depurarVerboso,
                                        lobj_concreteBuilderClass +
                                        " TODOS OS AGENTES JA' REGISTADOS" );
                  U_Depurador.escrever( kint_depurarInfo,
                                        "Agentes " + lobj_concreteBuilderClass+ " | Total = " +
                                        ie_initiator.obterNumeroThreads( lobj_concreteBuilderClass) );
               }
               else
               {
                  U_LeitorEscritorLinha.escrever( "*** Registo do Agente sem sucesso ***" );
                  U_LeitorEscritorLinha.escrever( "*** Motivo: " + lstr_motivo + " ***" );
                  System.exit( 1 );
               }
            }

            U_Depurador.escrever( kint_depurarVerboso,
                                  "--- DEPOIS: solicitar inicio Thread: " +
                                  U_Geral.paraString( lint_cicloThread ) );
            U_Depurador.escrever( kint_depurarVerboso, "" );
         }
      }
   }


   // PTS: A Thread e' iniciada imediatamente
   // (em construirRegistarCriarThreadParaCadaAgente)
   // - vai processando enquanto as outras fazem a conexao!!...
   private void iniciarThreadParaCadaAgente()
   {
   /*
      for( int lint_ciclo = 0;
           lint_ciclo < ilst_threads.length;
           lint_ciclo++ )
      {
         // Iniciar cada uma das "Threads" (cada um dos Agentes)
         for( Iterator litr_agentes = ilst_threads[ lint_ciclo ].iterator();
              litr_agentes.hasNext();
              ( ( Thread )litr_agentes.next() ).start() );
      }
   */
   }


   private void definirLigacoes()
   {
      ie_initiator = new E_Initiator();
      ib_initiator = new B_Initiator();
   }
}
