/* PTS */

package rCr.initiator;

import rCr.util.U_Geral;
import rCr.util.U_LeitorEscritorLinha;
import rCr.util.U_Depurador;
import rCr.util.I_Depurador;
import rCr.util.U_PreCondicao;
import rCr.util.X_PreCondicao;
import java.util.Hashtable;
import java.util.List;
import java.util.ArrayList;
import java.util.Enumeration;

/**
 *  @author PTS
 *  @version 0.0
 *  @since SET-2001
 */
public class E_Initiator implements I_Config, I_Depurador
{
   
   /**
   * Classe Iniciadora - tipo de agente (p. 35 manual)
   * 1 - Civilian
   * 2 - FireBrigade
   * 4 - FireStation
   * 8 - AmbulanceTeam
   * 16 - AmbulanceCenter
   * 32 - PoliceForce
   * 64 - PoliceOffice
   */
   private Hashtable ihsh_classeIniciadoraNumeroThreads = new Hashtable( 10 );
   
   /**
 *     Nome da mquina que executa o processo servidor
    */
   private String istr_nomeNoRemoto = kstr_nomeNoRemoto;
   
   /**
   * Porto onde o processo servidor "escuta"
   */
   private int iint_portoProcessoRemoto = kint_portoProcessoRemoto;

   /**
 *     Classe que implementa o mecanismo basico de comunicacao
 *     com o mundo exterior a este processo
    */
   private String istr_classeComunicadora = kstr_classeComunicadoraOmissao;
   
   public E_Initiator() 
   {
   }


   public boolean registarOpcoes( String[] astr_opcoes )
   {
      int lint_numeroOpcaoCorrente = 0;
      int lint_numeroOpcaoAnterior = 0;
      boolean lbln_opcaoValida = true;
      boolean lbln_apresentarInfoAjuda = false;

      // Percorrer os argumentos esperados na linha de execucao
      //=======================================================
      while( lbln_opcaoValida
             &&
             ( lint_numeroOpcaoCorrente < astr_opcoes.length )
             &&
             ( ! lbln_apresentarInfoAjuda ) )
      {
         // Caso seja necessario apresentar informacao de
         // ajuda e' desnecessario analisar os proximos argumentos
         lbln_apresentarInfoAjuda = apresentarInfoAjuda( astr_opcoes, lint_numeroOpcaoCorrente );
         if( ! lbln_apresentarInfoAjuda )
         {
            lint_numeroOpcaoCorrente = registarNomeNoRemoto( astr_opcoes, lint_numeroOpcaoCorrente );
            lint_numeroOpcaoCorrente = registarClasseIniciadoraNumeroThreads( astr_opcoes, lint_numeroOpcaoCorrente );
         }

         lbln_opcaoValida = ( lint_numeroOpcaoAnterior < lint_numeroOpcaoCorrente  );
         lint_numeroOpcaoAnterior = lint_numeroOpcaoCorrente;
      }
      
      // Apresentar info ajuda, tambem no caso em que nao se indica nenhuma classe iniciadora
      lbln_apresentarInfoAjuda = lbln_apresentarInfoAjuda || ( 0 == obterNumeroClassesIniciadoras() );

      return lbln_opcaoValida && ( ! lbln_apresentarInfoAjuda );
   }
   
   /**
 *     @roseuid 3BD05E3300D9
    */
   public int obterNumeroClassesIniciadoras() 
   {
      return ihsh_classeIniciadoraNumeroThreads.size();
   }
   
   /**
 *     @roseuid 3BCEF2B7004B
    */
   public Enumeration obterClassesIniciadoras() 
   {
      return ihsh_classeIniciadoraNumeroThreads.keys();
   }
   
   /**
 *     @roseuid 3BB1BC760346
    */
   public int obterNumeroThreads(Object aobj_objecto) 
   {
      int lint_numeroThreads = kint_numeroThreads;
      try
      {
         lint_numeroThreads = U_Geral.paraInt( ( Integer )ihsh_classeIniciadoraNumeroThreads.get( aobj_objecto ) );
      }
      catch( IndexOutOfBoundsException lexc_excepcao )
      {
         U_LeitorEscritorLinha.escrever( lexc_excepcao );
         return lint_numeroThreads;
      }

      U_Depurador.escrever( kint_depurarVerboso,
                            U_Geral.paraNomeClasse( this ),
                            "Numero Threads: " + U_Geral.paraString( lint_numeroThreads ) );

      return lint_numeroThreads;
   }


   public void atribuirNumeroThreads( Object aobj_classeIniciadora, int aint_numeroThreads )
   {
      ihsh_classeIniciadoraNumeroThreads.put( aobj_classeIniciadora,
                                              U_Geral.paraInteger( aint_numeroThreads ) );
   }


   /**
 *     @roseuid 3BB21769011E
    */
   public String obterNomeNoRemoto() 
   {
      return istr_nomeNoRemoto;
   }
   
   /**
 *     @roseuid 3BB34CE20386
    */
   public int obterPortoProcessoRemoto() 
   {
      return iint_portoProcessoRemoto;
   }


   /**
   * Objecto que implementa o mecanismo basico de comunicacao
   * com o mundo exterior a este processo
   */
   public String obterNomeClasseComunicadora()
   {
      return istr_classeComunicadora;
   }

   private int registarNomeNoRemoto( String[] astr_opcoes, int aint_numeroOpcao )
   {
      // Pre-condicao:
      // opcao em analise ser inferior que a dimensao do vector
      try
      {
         U_PreCondicao.assercaoVerdadeira(
            aint_numeroOpcao < astr_opcoes.length,
            this, "registarNomeNoRemoto" );
      }
      catch( X_PreCondicao lexc_preCondicao )
      {
         return aint_numeroOpcao;
      }

      // Opcao: -h para definicao da no' que executa o processo remoto
      if( U_Geral.igual( "-h", astr_opcoes[ aint_numeroOpcao ] )
	      /*&&
          ! Character.isDigit( astr_opcoes[ aint_numeroOpcao + 1 ].charAt( 0 ) )*/ )
      {
         istr_nomeNoRemoto = astr_opcoes[ aint_numeroOpcao + 1 ];
         aint_numeroOpcao = aint_numeroOpcao + 2;
      }

      return aint_numeroOpcao;
   }

   private int registarClasseIniciadoraNumeroThreads(String[] astr_opcoes, int aint_numeroOpcao) 
   {
      // Pre-condicao:
      // opcao em analise ser inferior que a dimensao do vector
      try
      {
         U_PreCondicao.assercaoVerdadeira(
            aint_numeroOpcao < astr_opcoes.length,
            this, "registarClasseIniciadoraNumeroThreads" );
      }
      catch( X_PreCondicao lexc_preCondicao )
      {
         return aint_numeroOpcao;
      }

      /** Opcao: indicacao de qual o tipo de agente a executar
      *
      * os Agentes "tipo humano"
      * ------------------------
      * -CV: Civilian
      * -FB: FireBrigade
      * -AT: AmbulanceTeam
      * -PF: PoliceForce
      * 
      * os Centros de Comando
      * ---------------------
      * -AC: AmbulanceCenter
      * -FS: FireStation
      * -PO: PoliceOffice
      */
      // Os Agentes "tipo humano"
      aint_numeroOpcao = registarParNomeValor( kstr_civilian, "-CV", astr_opcoes, aint_numeroOpcao );
      aint_numeroOpcao = registarParNomeValor( kstr_fireBrigade, "-FB", astr_opcoes, aint_numeroOpcao );
      aint_numeroOpcao = registarParNomeValor( kstr_ambulanceTeam, "-AT", astr_opcoes, aint_numeroOpcao );
      aint_numeroOpcao = registarParNomeValor( kstr_policeForce, "-PF", astr_opcoes, aint_numeroOpcao );

      // Os Centros de Comando
      aint_numeroOpcao = registarParNomeValor( kstr_fireStation, "-FS", astr_opcoes, aint_numeroOpcao );
      aint_numeroOpcao = registarParNomeValor( kstr_ambulanceCenter, "-AC", astr_opcoes, aint_numeroOpcao );
      aint_numeroOpcao = registarParNomeValor( kstr_policeOffice, "-PO", astr_opcoes, aint_numeroOpcao );

      return aint_numeroOpcao;
   }


   private int registarParNomeValor(String astr_nome, String astr_opcaoValida, String[] astr_opcoes, int aint_numeroOpcao) 
   {
      // Pre-condicao:
      // opcao em analise ser inferior que a dimensao do vector
      try
      {
         U_PreCondicao.assercaoVerdadeira(
            aint_numeroOpcao < astr_opcoes.length,
            this, "registarParNomeValor" );
      }
      catch( X_PreCondicao lexc_preCondicao )
      {
         return aint_numeroOpcao;
      }

      // Se a opcao for valida,
      // construir lista: [astr_nome, valor]
      // astr_nome: nome associado 'a "astr_opcaoValida"
      // valor: numero indicado a seguir 'a opcao (por omissao tera' o valor 1)
      if( U_Geral.igual( astr_opcaoValida.toUpperCase(), astr_opcoes[ aint_numeroOpcao ].toUpperCase() ) )
      {
         // Construir lista
         List llst_parNomeValor = new ArrayList( 2 );

         // Adicionar 'a lista o nome "astr_nome"
         llst_parNomeValor.add( astr_nome );
         aint_numeroOpcao = aint_numeroOpcao + 1;

         // Adicionar 'a lista o valor "lint_valor" associado a "astr_nome", ou o valor
         // definido por "I_Config.kint_numeroThreads", caso nao se encontre nenhum valor associado
         int lint_valor = kint_numeroThreads;
         if( aint_numeroOpcao < astr_opcoes.length )
         {
            if( Character.isDigit( astr_opcoes[ aint_numeroOpcao ].charAt( 0 ) ) )
            {
               lint_valor = U_Geral.paraInt( astr_opcoes[ aint_numeroOpcao ] );
               aint_numeroOpcao = aint_numeroOpcao + 1;
            }
         }

         // Adicionar 'a lista o valor "lint_valor"
         llst_parNomeValor.add( U_Geral.paraInteger( lint_valor ) );

         // Adicionar 'a variavel de instancia a lista "llst_parNomeValor" aqui construida
         //===============================================================================
         ihsh_classeIniciadoraNumeroThreads.put( llst_parNomeValor.get( 0 ), llst_parNomeValor.get( 1 ) );
      }

      return aint_numeroOpcao;
   }


   private boolean apresentarInfoAjuda(String[] astr_opcoes, int aint_numeroOpcao) 
   {
      // Por omissao nao apresenta informacao de ajuda
      boolean lbln_apresentarInfoAjuda = false;

      // Pre-condicao:
      // opcao em analise ser inferior que a dimensao do vector
      try
      {
         U_PreCondicao.assercaoVerdadeira(
            aint_numeroOpcao < astr_opcoes.length,
            this, "apresentarInfoAjuda" );
      }
      catch( X_PreCondicao lexc_preCondicao )
      {
         return lbln_apresentarInfoAjuda;
      }

      // Opcao: ? para apresentar informacao de ajuda
      if( U_Geral.igual( "?", astr_opcoes[ aint_numeroOpcao ] ) )
      {
         lbln_apresentarInfoAjuda = true;;
      }

      return lbln_apresentarInfoAjuda;
   }
}
