package rCr.util;

import java.io.IOException;
import java.io.ByteArrayOutputStream;
import java.io.DataOutputStream;
import java.io.ObjectOutputStream;


public class U_ListaDigitoBinario
{
   private ByteArrayOutputStream ibao_vectorBytes = new ByteArrayOutputStream();

   /**
    * A DataOutputStream lets an application
    * write primitive Java data types to an
    * output stream in a portable way 
    */
   private DataOutputStream idos_escritorTipoPrimitivo = new DataOutputStream( ibao_vectorBytes );

   /**
    * An ObjectOutputStream writes primitive data types
    * and graphs of Java objects to an OutputStream.
    * 
    * Only objects that support the
    * java.io.Serializable interface can be written to streams.
    * The class of each serializable object is encoded including
    * the class name and signature of the class,
    * the values of the object's fields and arrays,
    * and the closure of any other objects referenced
    * from the initial objects.
    * 
    * The default serialization mechanism for an object
    * writes the class of the object,
    * the class signature,
    * and the values of all non-transient and non-static fields.
    * 
    * References to other objects (except in transient or
    * static fields) cause those objects to be written also.
    * Multiple references to a single object are encoded
    * using a reference sharing mechanism so that
    * graphs of objects can be restored to the same shape
    * as when the original was written.
    * 
    * Default serialization for a class can be overridden
    * using the writeObject and the readObject methods.
    * 
    * Objects referenced by this object are written transitively
    * so that a complete equivalent graph of objects can be
    * reconstructed by an ObjectInputStream.
    */
   private ObjectOutputStream idos_escritorObjecto = null;


   public U_ListaDigitoBinario adicionar( short ashr_valor )
   {
      try
      {
         idos_escritorTipoPrimitivo.writeShort( ashr_valor );
         idos_escritorTipoPrimitivo.flush();
      }
      catch( IOException lexc_excepcao )
      {
         U_Depurador.escrever( lexc_excepcao );
      }

      return this;
   }


   public U_ListaDigitoBinario adicionar( int aint_valor )
   {
      try
      {
         idos_escritorTipoPrimitivo.writeInt( aint_valor );
         idos_escritorTipoPrimitivo.flush();
      }
      catch( IOException lexc_excepcao )
      {
         U_Depurador.escrever( lexc_excepcao );
      }

      return this;
   }


   public U_ListaDigitoBinario adicionar( String astr_valor )
   {
      try
      {
         idos_escritorTipoPrimitivo.writeBytes( astr_valor );
         idos_escritorTipoPrimitivo.flush();
      }
      catch( IOException lexc_excepcao )
      {
         U_Depurador.escrever( lexc_excepcao );
      }

      return this;
   }


   public U_ListaDigitoBinario adicionar( U_ListaDigitoBinario acl_valor )
   {
      try
      {
         idos_escritorTipoPrimitivo.write( acl_valor.obterVectorBytes(),
                                         0, acl_valor.obterDimensao() );
         idos_escritorTipoPrimitivo.flush();
      }
      catch( IOException lexc_excepcao )
      {
         U_Depurador.escrever( lexc_excepcao );
      }

      return this;
   }


   public U_ListaDigitoBinario adicionar( Object acl_objecto )
   {
      try
      {
         if( null == idos_escritorObjecto )
            idos_escritorObjecto = new ObjectOutputStream( ibao_vectorBytes );

         idos_escritorObjecto.writeObject( acl_objecto );
         idos_escritorTipoPrimitivo.flush();
      }
      catch( IOException lexc_excepcao )
      {
         U_Depurador.escrever( lexc_excepcao );
      }

      return this;
   }


   public int obterDimensao()
   {
      return ibao_vectorBytes.size();
   }


   public byte[] obterVectorBytes()
   {
      return ibao_vectorBytes.toByteArray();
   }


   public void escrever()
   {
      byte[] lbyt_vector = ibao_vectorBytes.toByteArray();
      U_LeitorEscritorLinha.escreverBinario( lbyt_vector );
   }


/*
   public U_ListaDigitoBinario()
   {
      super();
      setDimensao( 0 );
   }


   public int getDimensao()
   {
      return dimensao;
   }


   public void setDimensao( int dimensao )
   {
      this.dimensao = dimensao;
   }


   public int size()
   {
      return getDimensao();
   }


   public U_ListaDigitoBinario adicionar( int aint_valor )
   {
      String lstr_representacaoDigitoBinario =
                  U_Geral.preencherEsquerda( Integer.toBinaryString( aint_valor ),
                                             U_Geral.kint_numeroBitsDoInt,
                                             '0' );

      int lint_dimensaoRepresentacaoDigitoBinario = lstr_representacaoDigitoBinario.length();
      int lint_dimensaoLista = getDimensao();
      for( int lint_ciclo = 0;
           lint_ciclo < lint_dimensaoRepresentacaoDigitoBinario;
           lint_ciclo++ )
      {
         if( '0' == lstr_representacaoDigitoBinario.charAt( lint_ciclo ) )
            clear( lint_ciclo + lint_dimensaoLista );
         else
            set( lint_ciclo + lint_dimensaoLista );
      }
      setDimensao( getDimensao() + lint_dimensaoRepresentacaoDigitoBinario );

      return this;
   }


   public void escrever()
   {
      System.out.println( "Conteudo de U_ListaDigitoBinario:" );
      for( int lint_ciclo = 0; lint_ciclo < getDimensao(); lint_ciclo++ )
         System.out.print( get( lint_ciclo ) ? '1' : '0' );
      System.out.println( "" );
   }
*/
}
